#!/usr/bin/env python

# runs after the job (and after the default post-filter)
from galaxy.tools.parameters import DataToolParameter


def validate_input(trans, error_map, param_values, page_param_map):
    dbkeys = set()
    data_param_names = set()
    data_params = 0
    for name, param in page_param_map.items():
        if isinstance(param, DataToolParameter):
            # for each dataset parameter
            if param_values.get(name, None) is not None:
                dbkeys.add(param_values[name].dbkey)
                data_params += 1
                # check meta data
                try:
                    param = param_values[name]
                    int(param.metadata.startCol)
                    int(param.metadata.endCol)
                    int(param.metadata.chromCol)
                    if param.metadata.strandCol is not None:
                        int(param.metadata.strandCol)
                except Exception:
                    error_msg = (
                        "The attributes of this dataset are not properly set. "
                        "Click the pencil icon in the history item to set the chrom, start, end and strand columns."
                    )
                    error_map[name] = error_msg
            data_param_names.add(name)
    if len(dbkeys) > 1:
        for name in data_param_names:
            error_map[name] = (
                "All datasets must belong to same genomic build, "
                "this dataset is linked to build '%s'" % param_values[name].dbkey
            )
    if data_params != len(data_param_names):
        for name in data_param_names:
            error_map[name] = "A dataset of the appropriate type is required"



"""The module describes the ``email`` error plugin."""

import logging

from galaxy.tools.errors import EmailErrorReporter
from galaxy.util import (
    string_as_bool,
    unicodify,
)
from . import ErrorPlugin

log = logging.getLogger(__name__)


class EmailPlugin(ErrorPlugin):
    """Send error report as an email"""

    plugin_type = "email"

    def __init__(self, **kwargs):
        self.app = kwargs["app"]
        self.redact_user_details_in_bugreport = self.app.config.redact_user_details_in_bugreport
        self.verbose = string_as_bool(kwargs.get("verbose", True))
        self.user_submission = string_as_bool(kwargs.get("user_submission", True))

    def submit_report(self, dataset, job, tool, **kwargs):
        """Send report as an email"""
        try:
            error_reporter = EmailErrorReporter(dataset.id, self.app)
            error_reporter.send_report(
                user=job.get_user(),
                email=kwargs.get("email", None),
                message=kwargs.get("message", None),
                redact_user_details_in_bugreport=self.redact_user_details_in_bugreport,
            )
            return ("Your error report has been sent", "success")
        except Exception as e:
            return (f"An error occurred sending the report by email: {unicodify(e)}", "danger")


__all__ = ("EmailPlugin",)



import json
import logging
from typing import (
    Any,
    Dict,
    List,
    Sequence,
)

import packaging.version

log = logging.getLogger(__name__)

SKIP_INPUT = object()


def json_wrap(inputs, input_values, profile, as_dict=None, handle_files="skip"):
    if as_dict is None:
        as_dict = {}

    for input in inputs.values():
        input_name = input.name
        value_wrapper = input_values[input_name]
        json_value = _json_wrap_input(input, value_wrapper, profile, handle_files=handle_files)
        if json_value is SKIP_INPUT:
            continue
        as_dict[input_name] = json_value
    return as_dict


def data_input_to_path(v):
    path = _cast_if_not_none(v, str)
    if path == "None":
        path = None
    return path


def data_collection_input_to_path(v):
    return v.all_paths


def data_collection_input_to_staging_path_and_source_path(
    v, invalid_chars: Sequence[str] = ("/",), include_collection_name: bool = False
) -> List[Dict[str, Any]]:
    staging_paths = v.get_all_staging_paths(
        invalid_chars=invalid_chars, include_collection_name=include_collection_name
    )
    source_paths = v.all_paths
    metadata_files = v.all_metadata_files
    return [
        {
            "staging_path": staging_path,
            "source_path": source_path,
            "metadata_files": [
                {"staging_path": f"{staging_path}.{mf[0]}", "source_path": mf[1]} for mf in metadata_files
            ],
        }
        for staging_path, source_path, metadata_files in zip(staging_paths, source_paths, metadata_files)
    ]


def data_input_to_staging_path_and_source_path(v, invalid_chars: Sequence[str] = ("/",)) -> Dict[str, Any]:
    staging_path = v.get_staging_path(invalid_chars=invalid_chars)
    return {
        "staging_path": staging_path,
        "source_path": data_input_to_path(v),
        "metadata_files": [
            {"staging_path": f"{staging_path}.{mf[0]}", "source_path": mf[1]} for mf in v.all_metadata_files
        ],
    }


def _json_wrap_input(input, value_wrapper, profile, handle_files="skip"):
    input_type = input.type

    if input_type == "repeat":
        repeat_job_value = []
        for d in value_wrapper:
            repeat_instance_job_value = {}
            json_wrap(input.inputs, d, profile, repeat_instance_job_value, handle_files=handle_files)
            repeat_job_value.append(repeat_instance_job_value)
        json_value = repeat_job_value
    elif input_type == "conditional":
        values = value_wrapper
        current = values["__current_case__"]
        conditional_job_value = {}
        json_wrap(input.cases[current].inputs, values, profile, conditional_job_value, handle_files=handle_files)
        test_param = input.test_param
        test_param_name = test_param.name
        test_value = _json_wrap_input(test_param, values[test_param_name], profile, handle_files=handle_files)
        conditional_job_value[test_param_name] = test_value
        json_value = conditional_job_value
    elif input_type == "section":
        values = value_wrapper
        section_job_value = {}
        json_wrap(input.inputs, values, profile, section_job_value, handle_files=handle_files)
        json_value = section_job_value
    elif input_type == "data" and input.multiple:
        if handle_files == "paths":
            json_value = [data_input_to_path(v) for v in value_wrapper]
        elif handle_files == "staging_path_and_source_path":
            json_value = [data_input_to_staging_path_and_source_path(v) for v in value_wrapper]
        elif handle_files == "skip":
            return SKIP_INPUT
        else:
            raise NotImplementedError()
    elif input_type == "data":
        if handle_files == "paths":
            json_value = data_input_to_path(value_wrapper)
        elif handle_files == "staging_path_and_source_path":
            json_value = data_input_to_staging_path_and_source_path(value_wrapper)
        elif handle_files == "skip":
            return SKIP_INPUT
        elif handle_files == "OBJECT":
            if value_wrapper:
                if isinstance(value_wrapper, list):
                    value_wrapper = value_wrapper[0]
                json_value = _hda_to_object(value_wrapper)
                if input.load_contents:
                    with open(str(value_wrapper), mode="rb") as fh:
                        json_value["contents"] = fh.read(input.load_contents).decode("utf-8", errors="replace")
                return json_value
            else:
                return None
        else:
            raise NotImplementedError()
    elif input_type == "data_collection":
        if handle_files == "skip":
            return SKIP_INPUT
        elif handle_files == "paths":
            return data_collection_input_to_path(value_wrapper)
        elif handle_files == "staging_path_and_source_path":
            return data_collection_input_to_staging_path_and_source_path(value_wrapper)
        raise NotImplementedError()
    elif input_type in ["text", "color", "hidden"]:
        if getattr(input, "optional", False) and value_wrapper is not None and value_wrapper.value is None:
            json_value = None
        else:
            json_value = _cast_if_not_none(value_wrapper, str)
    elif input_type == "float":
        json_value = _cast_if_not_none(value_wrapper, float, empty_to_none=True)
    elif input_type == "integer":
        json_value = _cast_if_not_none(value_wrapper, int, empty_to_none=True)
    elif input_type == "boolean":
        if input.optional and value_wrapper is not None and value_wrapper.value is None:
            json_value = None
        else:
            json_value = _cast_if_not_none(value_wrapper, bool, empty_to_none=input.optional)
    elif input_type == "select":
        if packaging.version.parse(str(profile)) < packaging.version.parse("20.05"):
            json_value = _cast_if_not_none(value_wrapper, str)
        else:
            if input.multiple:
                json_value = [str(_) for _ in _cast_if_not_none(value_wrapper.value, list)]
            else:
                json_value = _cast_if_not_none(value_wrapper.value, str)
    elif input_type == "data_column":
        # value is a SelectToolParameterWrapper()
        if input.multiple:
            json_value = [int(_) for _ in _cast_if_not_none(value_wrapper.value, list)]
        else:
            json_value = [_cast_if_not_none(value_wrapper.value, int)]
    elif input_type == "directory_uri":
        json_value = _cast_if_not_none(value_wrapper, str)
    else:
        raise NotImplementedError(f"input_type [{input_type}] not implemented")

    return json_value


def _hda_to_object(hda):
    if hda.extension == "expression.json":
        # We may have a null data value
        with open(str(hda)) as inp:
            try:
                rval = json.loads(inp.read(5))
                if rval is None:
                    return rval
            except Exception:
                pass
    hda_dict = hda.to_dict()
    metadata_dict = {}

    for key, value in hda_dict.items():
        if key.startswith("metadata_"):
            metadata_dict[key[len("metadata_") :]] = value

    return {
        "file_ext": hda_dict["file_ext"],
        "file_size": hda_dict["file_size"],
        "name": hda_dict["name"],
        "metadata": metadata_dict,
        "src": {"src": "hda", "id": hda.id},
    }


def _cast_if_not_none(value, cast_to, empty_to_none=False):
    # log.debug("value [%s], type[%s]" % (value, type(value)))
    if value is None or (empty_to_none and str(value) == ""):
        return None
    else:
        return cast_to(value)


__all__ = ("json_wrap",)



import json
import logging
import os
import shutil
import sqlite3
import tempfile
import zlib
from threading import Lock

from sqlitedict import SqliteDict

from galaxy.util import unicodify
from galaxy.util.hash_util import md5_hash_file

log = logging.getLogger(__name__)

CURRENT_TOOL_CACHE_VERSION = 0


def encoder(obj):
    return sqlite3.Binary(zlib.compress(json.dumps(obj).encode("utf-8")))


def decoder(obj):
    return json.loads(zlib.decompress(bytes(obj)).decode("utf-8"))


class ToolDocumentCache:
    def __init__(self, cache_dir):
        self.cache_dir = cache_dir
        if not os.path.exists(self.cache_dir):
            os.makedirs(self.cache_dir)
        self.cache_file = os.path.join(self.cache_dir, "cache.sqlite")
        self.writeable_cache_file = None
        self._cache = None
        self.disabled = False
        self._get_cache(create_if_necessary=True)

    def close(self):
        self._cache and self._cache.close()

    def _get_cache(self, flag="r", create_if_necessary=False):
        try:
            if create_if_necessary and not os.path.exists(self.cache_file):
                # Create database if necessary using 'c' flag
                self._cache = SqliteDict(self.cache_file, flag="c", encode=encoder, decode=decoder, autocommit=False)
                if flag == "r":
                    self._cache.flag = flag
            else:
                cache_file = self.writeable_cache_file.name if self.writeable_cache_file else self.cache_file
                self._cache = SqliteDict(cache_file, flag=flag, encode=encoder, decode=decoder, autocommit=False)
        except sqlite3.OperationalError:
            log.warning("Tool document cache unavailable")
            self._cache = None
            self.disabled = True

    @property
    def cache_file_is_writeable(self):
        return os.access(self.cache_file, os.W_OK)

    def reopen_ro(self):
        self._get_cache(flag="r")
        self.writeable_cache_file = None

    def get(self, config_file):
        try:
            tool_document = self._cache.get(config_file)
        except sqlite3.OperationalError:
            log.debug("Tool document cache unavailable")
            return None
        if not tool_document:
            return None
        if tool_document.get("tool_cache_version") != CURRENT_TOOL_CACHE_VERSION:
            return None
        if self.cache_file_is_writeable:
            for path, modtime in tool_document["paths_and_modtimes"].items():
                if os.path.getmtime(path) != modtime:
                    return None
        return tool_document

    def _make_writable(self):
        if not self.writeable_cache_file:
            self.writeable_cache_file = tempfile.NamedTemporaryFile(
                dir=self.cache_dir, suffix="cache.sqlite.tmp", delete=False
            )
            if os.path.exists(self.cache_file):
                shutil.copy(self.cache_file, self.writeable_cache_file.name)
            self._get_cache(flag="c")

    def persist(self):
        if self.writeable_cache_file:
            self._cache.commit()
            os.rename(self.writeable_cache_file.name, self.cache_file)
            self.reopen_ro()

    def set(self, config_file, tool_source):
        try:
            if self.cache_file_is_writeable:
                self._make_writable()
                to_persist = {
                    "document": tool_source.to_string(),
                    "macro_paths": tool_source.macro_paths,
                    "paths_and_modtimes": tool_source.paths_and_modtimes(),
                    "tool_cache_version": CURRENT_TOOL_CACHE_VERSION,
                }
                try:
                    self._cache[config_file] = to_persist
                except RuntimeError:
                    log.debug("Tool document cache not writeable")
        except sqlite3.OperationalError:
            log.debug("Tool document cache unavailable")

    def delete(self, config_file):
        if self.cache_file_is_writeable:
            self._make_writable()
            try:
                del self._cache[config_file]
            except (KeyError, RuntimeError):
                pass

    def __del__(self):
        if self.writeable_cache_file:
            try:
                os.unlink(self.writeable_cache_file.name)
            except Exception:
                pass


class ToolCache:
    """
    Cache tool definitions to allow quickly reloading the whole
    toolbox.
    """

    def __init__(self):
        self._lock = Lock()
        self._hash_by_tool_paths = {}
        self._tools_by_path = {}
        self._tool_paths_by_id = {}
        self._macro_paths_by_id = {}
        self._new_tool_ids = set()
        self._removed_tool_ids = set()
        self._removed_tools_by_path = {}
        self._hashes_initialized = False

    def assert_hashes_initialized(self):
        if not self._hashes_initialized:
            for tool_hash in self._hash_by_tool_paths.values():
                tool_hash.hash
            self._hashes_initialized = True

    def cleanup(self):
        """
        Remove uninstalled tools from tool cache if they are not on disk anymore or if their content has changed.

        Returns list of tool_ids that have been removed.
        """
        removed_tool_ids = []
        try:
            with self._lock:
                persist_tool_document_cache = False
                paths_to_cleanup = {
                    (path, tool) for path, tool in self._tools_by_path.items() if self._should_cleanup(path)
                }
                for config_filename, tool in paths_to_cleanup:
                    tool.remove_from_cache()
                    persist_tool_document_cache = True
                    del self._hash_by_tool_paths[config_filename]
                    if os.path.exists(config_filename):
                        # This tool has probably been broken while editing on disk
                        # We record it here, so that we can recover it
                        self._removed_tools_by_path[config_filename] = self._tools_by_path[config_filename]
                    del self._tools_by_path[config_filename]
                    tool_ids = tool.all_ids
                    for tool_id in tool_ids:
                        if tool_id in self._tool_paths_by_id:
                            del self._tool_paths_by_id[tool_id]
                    removed_tool_ids.extend(tool_ids)
                for tool_id in removed_tool_ids:
                    self._removed_tool_ids.add(tool_id)
                    if tool_id in self._new_tool_ids:
                        self._new_tool_ids.remove(tool_id)
                if persist_tool_document_cache:
                    tool.app.toolbox.persist_cache()
        except Exception as e:
            log.debug("Exception while checking tools to remove from cache: %s", unicodify(e))
            # If by chance the file is being removed while calculating the hash or modtime
            # we don't want the thread to die.
        if removed_tool_ids:
            log.debug(f"Removed the following tools from cache: {removed_tool_ids}")
        return removed_tool_ids

    def _should_cleanup(self, config_filename):
        """Return True if `config_filename` does not exist or if modtime and hash have changes, else return False."""
        try:
            new_mtime = os.path.getmtime(config_filename)
            tool_hash = self._hash_by_tool_paths.get(config_filename)
            if tool_hash.modtime < new_mtime:
                if md5_hash_file(config_filename) != tool_hash.hash:
                    return True
            tool = self._tools_by_path[config_filename]
            for macro_path in tool._macro_paths:
                new_mtime = os.path.getmtime(macro_path)
                if self._hash_by_tool_paths.get(macro_path).modtime < new_mtime:
                    return True
        except FileNotFoundError:
            return True
        return False

    def get_tool(self, config_filename):
        """Get the tool at `config_filename` from the cache if the tool is up to date."""
        return self._tools_by_path.get(config_filename, None)

    def get_removed_tool(self, config_filename):
        return self._removed_tools_by_path.get(config_filename)

    def get_tool_by_id(self, tool_id):
        """Get the tool with the id `tool_id` from the cache if the tool is up to date."""
        return self.get_tool(self._tool_paths_by_id.get(tool_id))

    def expire_tool(self, tool_id):
        with self._lock:
            if tool_id in self._tool_paths_by_id:
                config_filename = self._tool_paths_by_id[tool_id]
                del self._hash_by_tool_paths[config_filename]
                del self._tool_paths_by_id[tool_id]
                del self._tools_by_path[config_filename]
                if tool_id in self._new_tool_ids:
                    self._new_tool_ids.remove(tool_id)

    def cache_tool(self, config_filename, tool):
        tool_id = str(tool.id)
        # We defer hashing of the config file if we haven't called assert_hashes_initialized.
        # This allows startup to occur without having to read in and hash all tool and macro files
        lazy_hash = not self._hashes_initialized
        with self._lock:
            self._hash_by_tool_paths[config_filename] = ToolHash(config_filename, lazy_hash=lazy_hash)
            self._tool_paths_by_id[tool_id] = config_filename
            self._tools_by_path[config_filename] = tool
            self._new_tool_ids.add(tool_id)
            for macro_path in tool._macro_paths:
                self._hash_by_tool_paths[macro_path] = ToolHash(macro_path, lazy_hash=lazy_hash)
                if tool_id not in self._macro_paths_by_id:
                    self._macro_paths_by_id[tool_id] = {macro_path}
                else:
                    self._macro_paths_by_id[tool_id].add(macro_path)

    def reset_status(self):
        """
        Reset tracking of new and newly disabled tools.
        """
        with self._lock:
            self._new_tool_ids = set()
            self._removed_tool_ids = set()
            self._removed_tools_by_path = {}


class ToolHash:
    def __init__(self, path, modtime=None, lazy_hash=False):
        self.path = path
        self.modtime = modtime or os.path.getmtime(path)
        self._tool_hash = None
        if not lazy_hash:
            self.hash

    @property
    def hash(self):
        if self._tool_hash is None:
            self._tool_hash = md5_hash_file(self.path)
        return self._tool_hash



"""
Execute an external process to evaluate expressions for Galaxy jobs.

Galaxy should be importable on sys.path .
"""
import json
import logging
import os
import sys
import warnings

# insert *this* galaxy before all others on sys.path
sys.path.insert(1, os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir, os.pardir)))
warnings.filterwarnings("ignore", message=r"[\n.]DEPRECATION: Python 2", module="cwltool")

try:
    from cwl_utils import expression
except ImportError:
    expression = None  # type: ignore[assignment]

from galaxy.tools.expressions import evaluate

logging.basicConfig()
log = logging.getLogger(__name__)


def run(environment_path=None):
    if expression is None:
        raise Exception("Python library cwltool must be available to evaluate expressions.")

    if environment_path is None:
        environment_path = os.environ.get("GALAXY_EXPRESSION_INPUTS")
    with open(environment_path) as f:
        raw_inputs = json.load(f)

    outputs = raw_inputs["outputs"]
    inputs = raw_inputs.copy()
    del inputs["outputs"]
    result = evaluate(None, inputs)

    if "__error_message" in result:
        raise Exception(result["__error_message"])
    for output in outputs:
        path = output["path"]
        from_expression = output["from_expression"]
        # if it is just a simple value, short-cut all the interpolation
        # interpolate seems to fail with None values so this worksaround
        # that for now.
        if from_expression in result:
            output_value = result[from_expression]
        else:
            from_expression = f"$({from_expression})"
            output_value = expression.interpolate(from_expression, result)
        with open(path, "w") as f:
            json.dump(output_value, f)



from .evaluation import evaluate
from .script import (
    EXPRESSION_SCRIPT_CALL,
    EXPRESSION_SCRIPT_NAME,
    write_evalute_script,
)
from .util import find_engine

__all__ = (
    "evaluate",
    "EXPRESSION_SCRIPT_CALL",
    "EXPRESSION_SCRIPT_NAME",
    "find_engine",
    "write_evalute_script",
)



import json
from importlib import import_module
from uuid import UUID

from pydantic import BaseModel

SCHEMA_MODELS_PACKAGE_BASE = "galaxy.schema."


# https://stackoverflow.com/a/2020083
def fullname(o):
    klass = o.__class__
    module = klass.__module__
    if module == "builtins":
        return klass.__qualname__  # avoid outputs like 'builtins.str'
    return module + "." + klass.__qualname__


class SchemaEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, BaseModel):
            return {
                "__type__": "__pydantic_object__",
                "__class__": fullname(obj),
                "__object__": obj.dict(),
            }
        if isinstance(obj, UUID):
            return str(obj)
        else:
            return json.JSONEncoder.default(self, obj)


def schema_decoder(obj):
    if "__type__" in obj:
        if obj["__type__"] == "__pydantic_object__":
            clazz_str = obj["__class__"]
            assert (
                clazz_str.startswith(SCHEMA_MODELS_PACKAGE_BASE) and ".." not in clazz_str
            ), f"Invalid class str {clazz_str}"
            module_name, class_name = clazz_str.rsplit(".", 1)
            module = import_module(module_name)
            clazz = getattr(module, class_name)
            obj = clazz(**obj["__object__"])
            return obj

    return obj


# Encoder function
def schema_dumps(obj):
    return json.dumps(obj, cls=SchemaEncoder)


# Decoder function
def schema_loads(obj):
    return json.loads(obj, object_hook=schema_decoder)



"""
Data providers for genome visualizations.
"""

import abc
import itertools
import math
import os
import random
import re
import sys
from contextlib import contextmanager
from json import loads
from typing import (
    Any,
    Dict,
    IO,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)

import pysam
from bx.bbi.bigbed_file import BigBedFile
from bx.bbi.bigwig_file import BigWigFile
from bx.interval_index_file import Indexes

from galaxy.datatypes.interval import (
    Bed,
    Gff,
    Gtf,
)
from galaxy.datatypes.util.gff_util import (
    convert_gff_coords_to_bed,
    GFFFeature,
    GFFInterval,
    GFFReaderWrapper,
    parse_gff_attributes,
)
from galaxy.model import DatasetInstance
from galaxy.visualization.data_providers.basic import BaseDataProvider
from galaxy.visualization.data_providers.cigar import get_ref_based_read_seq_and_cigar

#
# Utility functions.
#

# pysam 0.16.0.1 emits logs containing the word 'Error', this can confuse the stdout/stderr checkers.
# Can be be removed once https://github.com/pysam-developers/pysam/issues/939 is resolved.
pysam.set_verbosity(0)

PAYLOAD_LIST_TYPE = List[Optional[Union[str, int, float, List[Tuple[int, int]]]]]


def float_nan(n):
    """
    Return None instead of NaN to pass jQuery 1.4's strict JSON
    """
    if n != n:  # NaN != NaN
        return None
    else:
        return float(n)


def get_bounds(reads, start_pos_index, end_pos_index):
    """
    Returns the minimum and maximum position for a set of reads.
    """
    max_low = sys.maxsize
    max_high = -sys.maxsize
    for read in reads:
        if read[start_pos_index] < max_low:
            max_low = read[start_pos_index]
        if read[end_pos_index] > max_high:
            max_high = read[end_pos_index]
    return max_low, max_high


def _convert_between_ucsc_and_ensemble_naming(chrom):
    """
    Convert between UCSC chromosome ('chr1') naming conventions and Ensembl
    naming conventions ('1')
    """
    if chrom.startswith("chr"):
        # Convert from UCSC to Ensembl
        return chrom[3:]
    else:
        # Convert from Ensembl to UCSC
        return "chr" + chrom


def _chrom_naming_matches(chrom1, chrom2):
    return (chrom1.startswith("chr") and chrom2.startswith("chr")) or (
        not chrom1.startswith("chr") and not chrom2.startswith("chr")
    )


class FeatureLocationIndexDataProvider(BaseDataProvider):
    """
    Reads/writes/queries feature location index (FLI) datasets.
    """

    def __init__(self, converted_dataset):
        self.converted_dataset = converted_dataset

    def get_data(self, query):
        # Init.
        textloc_file = open(self.converted_dataset.file_name)
        line_len = int(textloc_file.readline())
        file_len = os.path.getsize(self.converted_dataset.file_name)
        query = query.lower()

        # Find query in file using binary search.
        low = 0
        high = int(file_len / line_len)
        while low < high:
            mid: int = (low + high) // 2
            position = mid * line_len
            textloc_file.seek(position)

            # Compare line with query and update low, high.
            line = textloc_file.readline()
            if line < query:
                low = mid + 1
            else:
                high = mid

        # Need to move back one line because last line read may be included in
        # results.
        position = low * line_len
        textloc_file.seek(position)

        # At right point in file, generate hits.
        result = []
        while True:
            line = textloc_file.readline()
            if not line.startswith(query):
                break
            if line[-1:] == "\n":
                line = line[:-1]
            result.append(line.split()[1:])

        textloc_file.close()
        return result


class GenomeDataProvider(BaseDataProvider):
    """
    Base class for genome data providers. All genome providers use BED coordinate
    format (0-based, half-open coordinates) for both queries and returned data.
    """

    dataset_type: str

    # Mapping from column name to payload data; this mapping is used to create
    # filters. Key is column name, value is a dict with mandatory key 'index'
    # and optional key 'name'. E.g. this defines column 4:
    # col_name_data_attr_mapping = {4 : { index: 5, name: 'Score' } }
    col_name_data_attr_mapping: Dict[Union[str, int], Dict] = {}

    def __init__(
        self,
        converted_dataset=None,
        original_dataset=None,
        dependencies=None,
        error_max_vals="Only the first %i %s in this region are displayed.",
    ):
        super().__init__(
            converted_dataset=converted_dataset,
            original_dataset=original_dataset,
            dependencies=dependencies,
            error_max_vals=error_max_vals,
        )

    def write_data_to_file(self, regions, filename):
        """
        Write data in region defined by chrom, start, and end to a file.
        """
        raise Exception("Unimplemented Function")

    def valid_chroms(self):
        """
        Returns chroms/contigs that the dataset contains
        """
        return None  # by default

    def has_data(self, chrom, start, end, **kwargs):
        """
        Returns true if dataset has data in the specified genome window, false
        otherwise.
        """
        raise Exception("Unimplemented Function")

    @contextmanager
    def open_data_file(self):
        """
        Open data file for reading data.
        """
        raise Exception("Unimplemented Function")

    def get_iterator(self, data_file, chrom, start, end, **kwargs) -> Iterator[str]:
        """
        Returns an iterator that provides data in the region chrom:start-end
        """
        raise Exception("Unimplemented Function")

    def process_data(self, iterator, start_val=0, max_vals=None, **kwargs):
        """
        Process data from an iterator to a format that can be provided to client.
        """
        raise Exception("Unimplemented Function")

    def get_data(self, chrom=None, low=None, high=None, start_val=0, max_vals=sys.maxsize, **kwargs):
        """
        Returns data in region defined by chrom, start, and end. start_val and
        max_vals are used to denote the data to return: start_val is the first element to
        return and max_vals indicates the number of values to return.

        Return value must be a dictionary with the following attributes:
            dataset_type, data
        """
        start, end = int(low), int(high)
        with self.open_data_file() as data_file:
            iterator = self.get_iterator(data_file, chrom, start, end, **kwargs)
            data = self.process_data(iterator, start_val, max_vals, start=start, end=end, **kwargs)
        return data

    def get_genome_data(self, chroms_info, **kwargs):
        """
        Returns data for complete genome.
        """
        genome_data = []
        for chrom_info in chroms_info["chrom_info"]:
            chrom = chrom_info["chrom"]
            chrom_len = chrom_info["len"]
            chrom_data = self.get_data(chrom, 0, chrom_len, **kwargs)
            # FIXME: data providers probably should never return None.
            # Some data providers return None when there's no data, so
            # create a dummy dict if necessary.
            if not chrom_data:
                chrom_data = {"data": None}
            chrom_data["region"] = "%s:%i-%i" % (chrom, 0, chrom_len)
            genome_data.append(chrom_data)

        return {"data": genome_data, "dataset_type": self.dataset_type}

    def get_filters(self):
        """
        Returns filters for provider's data. Return value is a list of
        filters; each filter is a dictionary with the keys 'name', 'index', 'type'.
        NOTE: This method uses the original dataset's datatype and metadata to
        create the filters.
        """
        # Get column names.
        try:
            column_names = self.original_dataset.datatype.column_names  # type: ignore[attr-defined]
        except AttributeError:
            try:
                column_names = list(range(self.original_dataset.metadata.columns))
            except Exception:  # Give up
                return []

        # Dataset must have column types; if not, cannot create filters.
        try:
            column_types = self.original_dataset.metadata.column_types
        except AttributeError:
            return []

        # Create and return filters.
        filters = []
        if self.original_dataset.metadata.viz_filter_cols:
            for viz_col_index in self.original_dataset.metadata.viz_filter_cols:
                # Some columns are optional, so can't assume that a filter
                # column is in dataset.
                if viz_col_index >= len(column_names):
                    continue
                col_name = column_names[viz_col_index]
                # Make sure that column has a mapped index. If not, do not add filter.
                try:
                    attrs = self.col_name_data_attr_mapping[col_name]
                except KeyError:
                    continue
                filters.append({"name": attrs["name"], "type": column_types[viz_col_index], "index": attrs["index"]})
        return filters

    def get_default_max_vals(self):
        return 5000


#
# -- Base mixins and providers --
#


class FilterableMixin:
    original_dataset: DatasetInstance

    def get_filters(self):
        """Returns a dataset's filters."""
        # Get filters.
        # TODOs:
        # (a) might be useful to move this into each datatype's set_meta method;
        # (b) could look at first N lines to ensure GTF attribute types are consistent.
        #
        filters = []
        # HACK: first 8 fields are for drawing, so start filter column index at 9.
        filter_col = 8
        if isinstance(self.original_dataset.datatype, Gff):
            # Can filter by score and GTF attributes.
            filters = [
                {"name": "Score", "type": "number", "index": filter_col, "tool_id": "Filter1", "tool_exp_name": "c6"}
            ]
            filter_col += 1
            if isinstance(self.original_dataset.datatype, Gtf):
                # Create filters based on dataset metadata.
                for name, a_type in self.original_dataset.metadata.attribute_types.items():
                    if a_type in ["int", "float"]:
                        filters.append(
                            {
                                "name": name,
                                "type": "number",
                                "index": filter_col,
                                "tool_id": "gff_filter_by_attribute",
                                "tool_exp_name": name,
                            }
                        )
                        filter_col += 1
        elif isinstance(self.original_dataset.datatype, Bed):
            # Can filter by score column only.
            filters = [
                {"name": "Score", "type": "number", "index": filter_col, "tool_id": "Filter1", "tool_exp_name": "c5"}
            ]

        return filters


class TabixDataProvider(GenomeDataProvider, FilterableMixin):
    """
    Tabix index data provider for the Galaxy track browser.
    """

    dataset_type = "tabix"

    col_name_data_attr_mapping: Dict[Union[str, int], Dict] = {4: {"index": 4, "name": "Score"}}

    @contextmanager
    def open_data_file(self):
        # We create a symlink to the index file. This is
        # required until https://github.com/pysam-developers/pysam/pull/586 is merged.
        index_path = self.converted_dataset.file_name
        with pysam.TabixFile(self.dependencies["bgzip"].file_name, index=index_path) as f:
            yield f

    def get_iterator(self, data_file, chrom, start, end, **kwargs) -> Iterator[str]:
        # chrom must be a string, start/end integers.
        # in previous versions of pysam, unicode was accepted for chrom, but not in 8.4
        chrom = str(chrom)
        start = int(start)
        end = int(end)
        if end >= (2 << 29):
            end = 2 << 29 - 1  # Tabix-enforced maximum
        # Get iterator using either naming scheme.
        iterator: Iterator[str] = iter([])
        if chrom in data_file.contigs:
            iterator = data_file.fetch(reference=chrom, start=start, end=end)
        else:
            # Try alternative naming scheme.
            chrom = _convert_between_ucsc_and_ensemble_naming(chrom)
            if chrom in data_file.contigs:
                iterator = data_file.fetch(reference=chrom, start=start, end=end)

        return iterator

    def write_data_to_file(self, regions, filename):
        with self.open_data_file() as data_file, open(filename, "w") as out:
            for region in regions:
                # Write data in region.
                iterator = self.get_iterator(data_file, region.chrom, region.start, region.end)
                for line in iterator:
                    out.write(f"{line}\n")


#
# -- Interval data providers --
#


class IntervalDataProvider(GenomeDataProvider):
    """
    Processes interval data from native format to payload format.

    Payload format: [ uid (offset), start, end, name, strand, thick_start, thick_end, blocks ]
    """

    dataset_type = "interval_index"

    def get_iterator(self, data_file, chrom, start, end, **kwargs):
        raise Exception("Unimplemented Function")

    def process_data(self, iterator, start_val=0, max_vals=None, **kwargs):
        """
        Provides
        """
        # Build data to return. Payload format is:
        # [ <guid/offset>, <start>, <end>, <name>, <strand> ]
        #
        # First three entries are mandatory, others are optional.
        #
        filter_cols = loads(kwargs.get("filter_cols", "[]"))
        no_detail = "no_detail" in kwargs
        rval = []
        message = None

        # Subtract one b/c columns are 1-based but indices are 0-based.
        def col_fn(col):
            return None if col is None else col - 1

        start_col = self.original_dataset.metadata.startCol - 1
        end_col = self.original_dataset.metadata.endCol - 1
        strand_col = col_fn(self.original_dataset.metadata.strandCol)
        name_col = col_fn(self.original_dataset.metadata.nameCol)
        for count, line in enumerate(iterator):
            if count < start_val:
                continue
            if max_vals and count - start_val >= max_vals:
                message = self.error_max_vals % (max_vals, "features")
                break

            feature = line.split()
            length = len(feature)
            # Unique id is just a hash of the line
            payload: PAYLOAD_LIST_TYPE = [hash(line), int(feature[start_col]), int(feature[end_col])]

            if no_detail:
                rval.append(payload)
                continue

            # Name, strand.
            if name_col:
                payload.append(feature[name_col])
            if strand_col:
                # Put empty name as placeholder.
                if not name_col:
                    payload.append("")
                payload.append(feature[strand_col])

            # Score (filter data)
            if length >= 5 and filter_cols and filter_cols[0] == "Score":
                try:
                    payload.append(float(feature[4]))
                except Exception:
                    payload.append(feature[4])

            rval.append(payload)

        return {"data": rval, "message": message}

    def write_data_to_file(self, regions, filename):
        raise Exception("Unimplemented Function")


class IntervalTabixDataProvider(TabixDataProvider, IntervalDataProvider):
    """
    Provides data from a BED file indexed via tabix.
    """


#
# -- BED data providers --
#


class BedDataProvider(GenomeDataProvider):
    """
    Processes BED data from native format to payload format.

    Payload format: [ uid (offset), start, end, name, strand, thick_start, thick_end, blocks ]
    """

    dataset_type = "interval_index"

    def get_iterator(self, data_file, chrom, start, end, **kwargs):
        raise Exception("Unimplemented Method")

    def process_data(self, iterator, start_val=0, max_vals=None, **kwargs):
        """
        Provides
        """
        # Build data to return. Payload format is:
        # [ <guid/offset>, <start>, <end>, <name>, <strand>, <thick_start>,
        #   <thick_end>, <blocks> ]
        #
        # First three entries are mandatory, others are optional.
        #
        filter_cols = loads(kwargs.get("filter_cols", "[]"))
        no_detail = "no_detail" in kwargs
        rval = []
        message = None
        for count, line in enumerate(iterator):
            if count < start_val:
                continue
            if max_vals and count - start_val >= max_vals:
                message = self.error_max_vals % (max_vals, "features")
                break
            # TODO: can we use column metadata to fill out payload?
            # TODO: use function to set payload data

            feature = line.split()
            length = len(feature)
            # Unique id is just a hash of the line
            payload: PAYLOAD_LIST_TYPE = [hash(line), int(feature[1]), int(feature[2])]

            if no_detail:
                rval.append(payload)
                continue

            # Name, strand, thick start, thick end.
            if length >= 4:
                payload.append(feature[3])
            if length >= 6:
                payload.append(feature[5])
            if length >= 8:
                payload.append(int(feature[6]))
                payload.append(int(feature[7]))

            # Blocks.
            if length >= 12:
                block_sizes = [int(n) for n in feature[10].split(",") if n != ""]
                block_starts = [int(n) for n in feature[11].split(",") if n != ""]
                blocks = list(zip(block_sizes, block_starts))
                payload.append(
                    [(int(feature[1]) + block[1], int(feature[1]) + block[1] + block[0]) for block in blocks]
                )

            # Score (filter data)
            if length >= 5 and filter_cols and filter_cols[0] == "Score":
                # If dataset doesn't have name/strand/thick start/thick end/blocks,
                # add placeholders. There should be 8 entries if all attributes
                # are present.
                payload.extend([None for i in range(8 - len(payload))])

                try:
                    payload.append(float(feature[4]))
                except Exception:
                    payload.append(feature[4])

            rval.append(payload)

        return {"data": rval, "dataset_type": self.dataset_type, "message": message}

    def write_data_to_file(self, regions, filename):
        with open(filename, "w") as out:
            for region in regions:
                # Write data in region.
                chrom = region.chrom
                start = region.start
                end = region.end
                with self.open_data_file() as data_file:
                    iterator = self.get_iterator(data_file, chrom, start, end)
                    for line in iterator:
                        out.write(f"{line}\n")


class BedTabixDataProvider(TabixDataProvider, BedDataProvider):
    """
    Provides data from a BED file indexed via tabix.
    """


class RawBedDataProvider(BedDataProvider):
    """
    Provide data from BED file.

    NOTE: this data provider does not use indices, and hence will be very slow
    for large datasets.
    """

    def get_iterator(self, data_file, chrom, start, end, **kwargs):
        # Read first line in order to match chrom naming format.
        line = data_file.readline()
        dataset_chrom = line.split()[0]
        if not _chrom_naming_matches(chrom, dataset_chrom):
            chrom = _convert_between_ucsc_and_ensemble_naming(chrom)
        # Undo read.
        data_file.seek(0)

        def line_filter_iter():
            with open(self.original_dataset.file_name) as data_file:
                for line in data_file:
                    if line.startswith("track") or line.startswith("browser"):
                        continue
                    feature = line.split()
                    feature_chrom = feature[0]
                    feature_start = int(feature[1])
                    feature_end = int(feature[2])
                    if (
                        (chrom is not None and feature_chrom != chrom)
                        or (start is not None and feature_start > end)
                        or (end is not None and feature_end < start)
                    ):
                        continue
                    yield line

        return line_filter_iter()


#
# -- VCF data providers --
#


class VcfDataProvider(GenomeDataProvider):
    """
    Abstract class that processes VCF data from native format to payload format.

    Payload format: An array of entries for each locus in the file. Each array
    has the following entries:

        1. GUID (unused)
        2. location (0-based)
        3. reference base(s)
        4. alternative base(s)
        5. quality score
        6. whether variant passed filter
        7. sample genotypes -- a single string with samples separated by commas; empty string
           denotes the reference genotype
        8. allele counts for each alternative

    """

    col_name_data_attr_mapping: Dict[Union[str, int], Dict] = {"Qual": {"index": 6, "name": "Qual"}}

    dataset_type = "variant"

    def process_data(self, iterator, start_val=0, max_vals=None, **kwargs):
        """
        Returns a dict with the following attributes::

            data - a list of variants with the format

            .. raw:: text

                [<guid>, <start>, <end>, <name>, cigar, seq]

            message - error/informative message

        """
        data = []
        message = None

        def get_mapping(ref, alt):
            """
            Returns ( offset, new_seq, cigar ) tuple that defines mapping of
            alt to ref. Cigar format is an array of [ op_index, length ] pairs
            where op_index is the 0-based index into the string "MIDNSHP=X"
            """

            cig_ops = "MIDNSHP=X"

            ref_len = len(ref)
            alt_len = len(alt)

            # Substitutions?
            if ref_len == alt_len:
                return 0, alt, [[cig_ops.find("M"), ref_len]]

            # Deletions?
            alt_in_ref_index = ref.find(alt)
            if alt_in_ref_index != -1:
                return alt_in_ref_index, ref[alt_in_ref_index + 1 :], [[cig_ops.find("D"), ref_len - alt_len]]

            # Insertions?
            ref_in_alt_index = alt.find(ref)
            if ref_in_alt_index != -1:
                return ref_in_alt_index, alt[ref_in_alt_index + 1 :], [[cig_ops.find("I"), alt_len - ref_len]]

        # Pack data.
        genotype_re = re.compile(r"/|\|")
        for count, line in enumerate(iterator):
            if count < start_val:
                continue
            if max_vals and count - start_val >= max_vals:
                message = self.error_max_vals % (max_vals, "features")
                break

            # Split line and aggregate data.
            feature = line.split()
            pos, c_id, ref, alt, qual, c_filter, info = feature[1:8]

            # Format and samples data are optional.
            samples_data = []
            if len(feature) > 8:
                samples_data = feature[9:]

            # VCF is 1-based but provided position is 0-based.
            pos = int(pos) - 1

            # FIXME: OK to skip?
            if alt == ".":
                count -= 1
                continue

            # Set up array to track allele counts.
            allele_counts = [0 for i in range(alt.count(",") + 1)]
            sample_gts = []

            if samples_data:
                # Process and pack samples' genotype and count alleles across samples.
                alleles_seen: Dict[int, bool] = {}
                has_alleles = False

                for sample in samples_data:
                    # Parse and count alleles.
                    genotype = sample.split(":")[0]
                    has_alleles = False
                    alleles_seen.clear()
                    for allele_str in genotype_re.split(genotype):
                        try:
                            # This may throw a ValueError if allele is missing.
                            allele = int(allele_str)

                            # Only count allele if it hasn't been seen yet.
                            if allele != 0 and allele not in alleles_seen:
                                allele_counts[allele - 1] += 1
                                alleles_seen[allele] = True
                                has_alleles = True
                        except ValueError:
                            pass

                    # If no alleles, use empty string as proxy.
                    if not has_alleles:
                        genotype = ""

                    sample_gts.append(genotype)
            else:
                # No samples, so set allele count and sample genotype manually.
                allele_counts = [1]
                sample_gts = ["1/1"]

            # Add locus data.
            locus_data = [-1, pos, c_id, ref, alt, qual, c_filter, ",".join(sample_gts)]
            locus_data.extend(allele_counts)
            data.append(locus_data)

        return {"data": data, "message": message}

    def write_data_to_file(self, regions, filename):
        out = open(filename, "w")
        with self.open_data_file() as data_file:
            for region in regions:
                # Write data in region.
                iterator = self.get_iterator(data_file, region.chrom, region.start, region.end)
                for line in iterator:
                    out.write(f"{line}\n")


class VcfTabixDataProvider(TabixDataProvider, VcfDataProvider):
    """
    Provides data from a VCF file indexed via tabix.
    """

    dataset_type = "variant"


class RawVcfDataProvider(VcfDataProvider):
    """
    Provide data from VCF file.

    NOTE: this data provider does not use indices, and hence will be very slow
    for large datasets.
    """

    @contextmanager
    def open_data_file(self):
        with open(self.original_dataset.file_name) as f:
            yield f

    def get_iterator(self, data_file, chrom, start, end, **kwargs):
        # Skip comments.
        line = None
        for line in data_file:
            if not line.startswith("#"):
                break

        # If last line is a comment, there are no data lines.
        if line.startswith("#"):
            return []

        # Match chrom naming format.
        if line:
            dataset_chrom = line.split()[0]
            if not _chrom_naming_matches(chrom, dataset_chrom):
                chrom = _convert_between_ucsc_and_ensemble_naming(chrom)

        def line_in_region(vcf_line, chrom, start, end):
            """Returns true if line is in region."""
            variant_chrom, variant_start = vcf_line.split()[0:2]
            # VCF format is 1-based.
            variant_start = int(variant_start) - 1
            return variant_chrom == chrom and variant_start >= start and variant_start <= end

        def line_filter_iter():
            """Yields lines in data that are in region chrom:start-end"""
            # Yield data line read above.
            if line_in_region(line, chrom, start, end):
                yield line

            # Search for and yield other data lines.
            for data_line in data_file:
                if line_in_region(data_line, chrom, start, end):
                    yield data_line

        return line_filter_iter()


class BamDataProvider(GenomeDataProvider, FilterableMixin):
    """
    Provides access to intervals from a sorted indexed BAM file. Coordinate
    data is reported in BED format: 0-based, half-open.
    """

    dataset_type = "bai"

    def get_filters(self):
        """
        Returns filters for dataset.
        """
        # HACK: first 7 fields are for drawing, so start filter column index at 7.
        filter_col = 7
        filters = []
        filters.append({"name": "Mapping Quality", "type": "number", "index": filter_col})
        return filters

    def write_data_to_file(self, regions, filename):
        """
        Write reads in regions to file.
        """

        # Open current BAM file using index.
        bamfile = pysam.AlignmentFile(
            self.original_dataset.file_name, mode="rb", index_filename=self.converted_dataset.file_name
        )

        # TODO: write headers as well?
        new_bamfile = pysam.AlignmentFile(filename, template=bamfile, mode="wb")

        for region in regions:
            # Write data from region.
            chrom = region.chrom
            start = region.start
            end = region.end

            try:
                data = bamfile.fetch(start=start, end=end, reference=chrom)
            except ValueError:
                # Try alternative chrom naming.
                chrom = _convert_between_ucsc_and_ensemble_naming(chrom)
                try:
                    data = bamfile.fetch(start=start, end=end, reference=chrom)
                except ValueError:
                    return None

            # Write reads in region.
            for read in data:
                new_bamfile.write(read)

        # Cleanup.
        new_bamfile.close()
        bamfile.close()

    @contextmanager
    def open_data_file(self):
        # Attempt to open the BAM file with index
        with pysam.AlignmentFile(
            self.original_dataset.file_name, mode="rb", index_filename=self.converted_dataset.file_name
        ) as f:
            yield f

    def get_iterator(self, data_file, chrom, start, end, **kwargs) -> Iterator[str]:
        """
        Returns an iterator that provides data in the region chrom:start-end
        """

        # Fetch and return data.
        chrom = str(chrom)
        start = int(start)
        end = int(end)
        try:
            data = data_file.fetch(start=start, end=end, reference=chrom)
        except ValueError:
            # Try alternative chrom naming.
            chrom = _convert_between_ucsc_and_ensemble_naming(chrom)
            try:
                data = data_file.fetch(start=start, end=end, reference=chrom)
            except ValueError:
                return iter([])
        return data

    def process_data(
        self,
        iterator,
        start_val=0,
        max_vals=None,
        ref_seq=None,
        iterator_type="nth",
        mean_depth=None,
        start=0,
        end=0,
        **kwargs,
    ):
        """
        Returns a dict with the following attributes::

            data - a list of reads with the format
                [<guid>, <start>, <end>, <name>, <read_1>, <read_2>, [empty], <mapq_scores>]

                where <read_1> has the format
                    [<start>, <end>, <cigar>, <strand>, <read_seq>]

                and <read_2> has the format
                    [<start>, <end>, <cigar>, <strand>, <read_seq>]

                Field 7 is empty so that mapq scores' location matches that in single-end reads.
                For single-end reads, read has format:
                    [<guid>, <start>, <end>, <name>, <cigar>, <strand>, <seq>, <mapq_score>]

                NOTE: read end and sequence data are not valid for reads outside of
                requested region and should not be used.

            max_low - lowest coordinate for the returned reads
            max_high - highest coordinate for the returned reads
            message - error/informative message
        """
        # No iterator indicates no reads.
        if iterator is None:
            return {"data": [], "message": None}

        #
        # Helper functions.
        #

        def decode_strand(read_flag, mask):
            """Decode strand from read flag."""

            strand_flag = read_flag & mask == 0
            if strand_flag:
                return "+"
            else:
                return "-"

        def _random_read_iterator(read_iterator, threshold):
            """
            An iterator that returns a random stream of reads from the read_iterator
            as well as corresponding pairs for returned reads.
            threshold is a value in [0,1] that denotes the percentage of reads
            to return.
            """
            for e in read_iterator:
                if e.qname in paired_pending or random.uniform(0, 1) <= threshold:
                    yield e

        def _nth_read_iterator(read_iterator, threshold):
            """
            An iterator that returns every nth read.
            """

            # Convert threshold to N for stepping through iterator.
            n = int(1 / threshold)
            return itertools.islice(read_iterator, None, None, n)

        # -- Choose iterator. --

        # Calculate threshold for non-sequential iterators based on mean_depth and read length.
        try:
            first_read = next(iterator)
        except StopIteration:
            # no reads.
            return {"data": [], "message": None, "max_low": start, "max_high": start}

        read_len = len(first_read.seq)
        num_reads = max((end - start) * mean_depth / float(read_len), 1)
        threshold = float(max_vals) / num_reads
        iterator = itertools.chain(iter([first_read]), iterator)

        # Use specified iterator type, save for when threshold is >= 1.
        # A threshold of >= 1 indicates all reads are to be returned, so no
        # sampling needed and seqential iterator will be used.
        if iterator_type == "sequential" or threshold >= 1:
            read_iterator = iterator
        elif iterator_type == "random":
            read_iterator = _random_read_iterator(iterator, threshold)
        elif iterator_type == "nth":
            read_iterator = _nth_read_iterator(iterator, threshold)

        #
        # Encode reads as list of lists.
        #
        results = []
        paired_pending: Dict[str, Dict[str, Any]] = {}
        unmapped = 0
        message = None
        count = 0
        for read in read_iterator:
            if count < start_val:
                continue
            if (count - start_val - unmapped) >= max_vals:
                message = self.error_max_vals % (max_vals, "reads")
                break

            # If not mapped, skip read.
            is_mapped = read.flag & 0x0004 == 0
            if not is_mapped:
                unmapped += 1
                continue

            qname = read.qname
            seq = read.seq
            strand = decode_strand(read.flag, 0x0010)
            if read.cigar is not None:
                read_len = sum(cig[1] for cig in read.cigar)  # Use cigar to determine length
            else:
                read_len = len(seq)  # If no cigar, just use sequence length

            if read.is_proper_pair:
                if qname in paired_pending:
                    # Found pair.
                    pair = paired_pending[qname]
                    results.append(
                        [
                            hash("%i_%s" % (pair["start"], qname)),
                            pair["start"],
                            read.pos + read_len,
                            qname,
                            [pair["start"], pair["end"], pair["cigar"], pair["strand"], pair["seq"]],
                            [read.pos, read.pos + read_len, read.cigar, strand, seq],
                            None,
                            [pair["mapq"], read.mapq],
                        ]
                    )
                    del paired_pending[qname]
                else:
                    # Insert first of pair.
                    paired_pending[qname] = {
                        "start": read.pos,
                        "end": read.pos + read_len,
                        "seq": seq,
                        "mate_start": read.mpos,
                        "rlen": read_len,
                        "strand": strand,
                        "cigar": read.cigar,
                        "mapq": read.mapq,
                    }
                    count += 1
            else:
                results.append(
                    [
                        hash("%i_%s" % (read.pos, qname)),
                        read.pos,
                        read.pos + read_len,
                        qname,
                        read.cigar,
                        strand,
                        read.seq,
                        read.mapq,
                    ]
                )
                count += 1

        # Take care of reads whose mates are out of range.
        for qname, read in paired_pending.items():
            if read["mate_start"] < read["start"]:
                # Mate is before read.
                read_start = read["mate_start"]
                read_end = read["end"]
                # Make read_1 start=end so that length is 0 b/c we don't know
                # read length.
                r1 = [read["mate_start"], read["mate_start"]]
                r2 = [read["start"], read["end"], read["cigar"], read["strand"], read["seq"]]
            else:
                # Mate is after read.
                read_start = read["start"]
                # Make read_2 start=end so that length is 0 b/c we don't know
                # read length. Hence, end of read is start of read_2.
                read_end = read["mate_start"]
                r1 = [read["start"], read["end"], read["cigar"], read["strand"], read["seq"]]
                r2 = [read["mate_start"], read["mate_start"]]

            results.append(
                [hash("%i_%s" % (read_start, qname)), read_start, read_end, qname, r1, r2, [read["mapq"], 125]]
            )

        # Clean up. TODO: is this needed? If so, we'll need a cleanup function after processing the data.
        # bamfile.close()

        def compress_seq_and_cigar(read, start_field, cigar_field, seq_field):
            """
            Use reference-based compression to compress read sequence and cigar.
            """
            read_seq, read_cigar = get_ref_based_read_seq_and_cigar(
                read[seq_field].upper(), read[start_field], ref_seq.sequence, ref_seq.start, read[cigar_field]
            )
            read[seq_field] = read_seq
            read[cigar_field] = read_cigar

        def convert_cigar(read, start_field, cigar_field, seq_field):
            """
            Convert read cigar from pysam format to string format.
            """
            cigar_ops = "MIDNSHP=X"
            read_cigar = ""
            for op_tuple in read[cigar_field]:
                read_cigar += "%i%s" % (op_tuple[1], cigar_ops[op_tuple[0]])
            read[cigar_field] = read_cigar

        # Choose method for processing reads. Use reference-based compression
        # if possible. Otherwise, convert cigar.
        if ref_seq:
            # Uppercase for easy comparison.
            ref_seq.sequence = ref_seq.sequence.upper()
            process_read = compress_seq_and_cigar
        else:
            process_read = convert_cigar

        # Process reads.
        for read in results:
            if isinstance(read[5], list):
                # Paired-end read.
                if len(read[4]) > 2:
                    process_read(read[4], 0, 2, 4)
                if len(read[5]) > 2:
                    process_read(read[5], 0, 2, 4)
            else:
                # Single-end read.
                process_read(read, 1, 4, 6)

        max_low, max_high = get_bounds(results, 1, 2)

        return {"data": results, "message": message, "max_low": max_low, "max_high": max_high}


class SamDataProvider(BamDataProvider):
    dataset_type = "bai"

    def __init__(self, converted_dataset=None, original_dataset=None, dependencies=None):
        """Create SamDataProvider."""
        super().__init__(
            converted_dataset=converted_dataset, original_dataset=original_dataset, dependencies=dependencies
        )

        # To use BamDataProvider, original dataset must be BAM and
        # converted dataset must be BAI. Use BAI from BAM metadata.
        if converted_dataset:
            self.original_dataset = converted_dataset
            self.converted_dataset = converted_dataset.metadata.bam_index


class BBIDataProvider(GenomeDataProvider):
    """
    BBI data provider for the Galaxy track browser.
    """

    dataset_type = "bigwig"

    @abc.abstractmethod
    def _get_dataset(self) -> Tuple[IO[bytes], Union[BigBedFile, BigWigFile]]:
        ...

    def valid_chroms(self):
        # No way to return this info as of now
        return None

    def has_data(self, chrom):
        f, bbi = self._get_dataset()
        all_dat = bbi.query(chrom, 0, 2147483647, 1) or bbi.query(
            _convert_between_ucsc_and_ensemble_naming(chrom), 0, 2147483647, 1
        )
        f.close()
        return all_dat is not None

    def get_data(self, chrom, start, end, start_val=0, max_vals=None, num_samples=1000, **kwargs):
        start = int(start)
        end = int(end)

        # Helper function for getting summary data regardless of chromosome
        # naming convention.
        def _summarize_bbi(bbi, chrom, start, end, num_points):
            return bbi.summarize(chrom, start, end, num_points) or bbi.summarize(
                _convert_between_ucsc_and_ensemble_naming(chrom), start, end, num_points
            )

        # Bigwig can be a standalone bigwig file, in which case we use
        # original_dataset, or coming from wig->bigwig conversion in
        # which we use converted_dataset
        f, bbi = self._get_dataset()

        # If stats requested, compute overall summary data for the range
        # start:endbut no reduced data. This is currently used by client
        # to determine the default range.
        if "stats" in kwargs:
            summary = _summarize_bbi(bbi, chrom, start, end, 1)
            f.close()

            min_val = 0
            max_val = 0
            mean = 0
            sd = 0.0
            if summary is not None:
                # Does the summary contain any defined values?
                valid_count = summary.valid_count[0]
                if summary.valid_count > 0:
                    # Compute $\mu \pm 2\sigma$ to provide an estimate for upper and lower
                    # bounds that contain ~95% of the data.
                    mean = summary.sum_data[0] / valid_count
                    var = max(summary.sum_squares[0] - mean, 0)  # Prevent variance underflow.
                    if valid_count > 1:
                        var /= valid_count - 1
                    sd = math.sqrt(var)
                    min_val = summary.min_val[0]
                    max_val = summary.max_val[0]

            return dict(data=dict(min=min_val, max=max_val, mean=mean, sd=sd))

        def summarize_region(bbi, chrom, start, end, num_points):
            """
            Returns results from summarizing a region using num_points.
            NOTE: num_points cannot be greater than end - start or BBI
            will return None for all positions.
            """
            result = []

            # Get summary; this samples at intervals of length
            # (end - start)/num_points -- i.e. drops any fractional component
            # of interval length.
            summary = _summarize_bbi(bbi, chrom, start, end, num_points)
            if summary:
                # mean = summary.sum_data / summary.valid_count

                # Standard deviation by bin, not yet used
                # var = summary.sum_squares - mean
                # var /= minimum( valid_count - 1, 1 )
                # sd = sqrt( var )

                pos = start
                step_size = (end - start) / num_points

                for i in range(num_points):
                    result.append((pos, float_nan(summary.sum_data[i] / summary.valid_count[i])))
                    pos += step_size

            return result

        # Approach is different depending on region size.
        num_samples = int(num_samples)
        if end - start < num_samples:
            # Get values for individual bases in region, including start and end.
            # To do this, need to increase end to next base and request number of points.
            num_points = end - start + 1
            end += 1
        else:
            #
            # The goal is to sample the region between start and end uniformly
            # using ~N (num_samples) data points. The challenge is that the size of
            # sampled intervals rarely is full bases, so sampling using N points
            # will leave the end of the region unsampled due to remainders for
            # each interval. To recitify this, a new N is calculated based on the
            # step size that covers as much of the region as possible.
            #
            # However, this still leaves some of the region unsampled. This
            # could be addressed by repeatedly sampling remainder using a
            # smaller and smaller step_size, but that would require iteratively
            # going to BBI, which could be time consuming.
            #

            # Start with N samples.
            num_points = num_samples
            step_size = (end - start) // num_points
            # Add additional points to sample in the remainder not covered by
            # the initial N samples.
            remainder_start = start + step_size * num_points
            additional_points = (end - remainder_start) // step_size
            num_points += additional_points

        result = summarize_region(bbi, chrom, start, end, num_points)

        # Cleanup and return.
        f.close()
        return {"data": result, "dataset_type": self.dataset_type}


class BigBedDataProvider(BBIDataProvider):
    def _get_dataset(self):
        # Nothing converts to bigBed so we don't consider converted dataset
        f = open(self.original_dataset.file_name, "rb")
        return f, BigBedFile(file=f)


class BigWigDataProvider(BBIDataProvider):
    """
    Provides data from BigWig files; position data is reported in 1-based
    coordinate system, i.e. wiggle format.
    """

    def _get_dataset(self):
        if self.converted_dataset is not None:
            f = open(self.converted_dataset.file_name, "rb")
        else:
            f = open(self.original_dataset.file_name, "rb")
        return f, BigWigFile(file=f)


class IntervalIndexDataProvider(GenomeDataProvider, FilterableMixin):
    """
    Interval index files used for GFF, Pileup files.
    """

    col_name_data_attr_mapping = {4: {"index": 4, "name": "Score"}}

    dataset_type = "interval_index"

    def write_data_to_file(self, regions, filename):
        index = Indexes(self.converted_dataset.file_name)
        with open(self.original_dataset.file_name) as source, open(filename, "w") as out:
            for region in regions:
                # Write data from region.
                chrom = region.chrom
                start = region.start
                end = region.end
                for _start, _end, offset in index.find(chrom, start, end):
                    source.seek(offset)

                    # HACK: write differently depending on original dataset format.
                    if self.original_dataset.ext not in ["gff", "gff3", "gtf"]:
                        line = source.readline()
                        out.write(line)
                    else:
                        reader = GFFReaderWrapper(source, fix_strand=True)
                        feature = next(reader)
                        for interval in feature.intervals:
                            out.write("\t".join(interval.fields) + "\n")

    @contextmanager
    def open_data_file(self):
        i = Indexes(self.converted_dataset.file_name)
        yield i

    def get_iterator(self, data_file, chrom, start, end, **kwargs) -> Iterator[str]:
        """
        Returns an iterator for data in data_file in chrom:start-end
        """
        if chrom not in data_file.indexes:
            # Try alternative naming.
            chrom = _convert_between_ucsc_and_ensemble_naming(chrom)

        return data_file.find(chrom, start, end)

    def process_data(self, iterator, start_val=0, max_vals=None, **kwargs):
        results = []
        message = None
        with open(self.original_dataset.file_name) as source:
            # Build data to return. Payload format is:
            # [ <guid/offset>, <start>, <end>, <name>, <score>, <strand>, <thick_start>,
            #   <thick_end>, <blocks> ]
            #
            # First three entries are mandatory, others are optional.
            filter_cols = loads(kwargs.get("filter_cols", "[]"))
            no_detail = "no_detail" in kwargs
            for count, val in enumerate(iterator):
                offset = val[2]
                if count < start_val:
                    continue
                if count - start_val >= max_vals:
                    message = self.error_max_vals % (max_vals, "features")
                    break
                source.seek(offset)
                # TODO: can we use column metadata to fill out payload?

                # GFF dataset.
                reader = GFFReaderWrapper(source, fix_strand=True)
                feature = next(reader)
                payload = package_gff_feature(feature, no_detail, filter_cols)
                payload.insert(0, offset)

                results.append(payload)

        return {"data": results, "message": message}


class RawGFFDataProvider(GenomeDataProvider):
    """
    Provide data from GFF file that has not been indexed.

    NOTE: this data provider does not use indices, and hence will be very slow
    for large datasets.
    """

    dataset_type = "interval_index"

    def get_iterator(self, data_file, chrom, start, end, **kwargs):
        """
        Returns an iterator that provides data in the region chrom:start-end as well as
        a file offset.
        """
        source = open(self.original_dataset.file_name)

        # Read first line in order to match chrom naming format.
        line = source.readline()

        # If line empty, assume file is empty and return empty iterator.
        if len(line) == 0:
            return iter([])

        # Determine chromosome naming format.
        dataset_chrom = line.split()[0]
        if not _chrom_naming_matches(chrom, dataset_chrom):
            chrom = _convert_between_ucsc_and_ensemble_naming(chrom)
        # Undo read.
        source.seek(0)

        def features_in_region_iter():
            offset = 0
            for feature in GFFReaderWrapper(source, fix_strand=True):
                # Only provide features that are in region.
                feature_start, feature_end = convert_gff_coords_to_bed([feature.start, feature.end])
                if feature.chrom == chrom and feature_end > start and feature_start < end:
                    yield feature, offset
                offset += feature.raw_size

        return features_in_region_iter()

    def process_data(self, iterator, start_val=0, max_vals=None, **kwargs):
        """
        Process data from an iterator to a format that can be provided to client.
        """
        filter_cols = loads(kwargs.get("filter_cols", "[]"))
        no_detail = "no_detail" in kwargs
        results = []
        message = None

        for count, (feature, offset) in enumerate(iterator):
            if count < start_val:
                continue
            if count - start_val >= max_vals:
                message = self.error_max_vals % (max_vals, "reads")
                break

            payload = package_gff_feature(feature, no_detail=no_detail, filter_cols=filter_cols)
            payload.insert(0, offset)
            results.append(payload)

        return {"data": results, "dataset_type": self.dataset_type, "message": message}


class GtfTabixDataProvider(TabixDataProvider):
    """
    Returns data from GTF datasets that are indexed via tabix.
    """

    def process_data(self, iterator, start_val=0, max_vals=None, **kwargs):
        # Loop through lines and group by transcript_id; each group is a feature.

        # TODO: extend this code or use code in gff_util to process GFF/3 as well
        # and then create a generic GFFDataProvider that can be used with both
        # raw and tabix datasets.
        features: Dict[str, List[GFFInterval]] = {}

        for line in iterator:
            line_attrs = parse_gff_attributes(line.split("\t")[8])
            transcript_id = line_attrs["transcript_id"]
            feature_list: List[GFFInterval]
            if transcript_id in features:
                feature_list = features[transcript_id]
            else:
                feature_list = []
                features[transcript_id] = feature_list
            feature_list.append(GFFInterval(None, line.split("\t")))

        # Process data.
        filter_cols = loads(kwargs.get("filter_cols", "[]"))
        no_detail = "no_detail" in kwargs
        results = []
        message = None

        for count, intervals in enumerate(features.values()):
            if count < start_val:
                continue
            if count - start_val >= max_vals:
                message = self.error_max_vals % (max_vals, "reads")
                break

            feature = GFFFeature(None, intervals=intervals)
            payload = package_gff_feature(feature, no_detail=no_detail, filter_cols=filter_cols)
            payload.insert(0, feature.intervals[0].attributes["transcript_id"])
            results.append(payload)

        return {"data": results, "message": message}


#
# -- ENCODE Peak data providers.
#


class ENCODEPeakDataProvider(GenomeDataProvider):
    """
    Abstract class that processes ENCODEPeak data from native format to payload format.

    Payload format: [ uid (offset), start, end, name, strand, thick_start, thick_end, blocks ]
    """

    def get_iterator(self, data_file, chrom, start, end, **kwargs):
        raise Exception("Unimplemented Method")

    def process_data(self, iterator, start_val=0, max_vals=None, **kwargs):
        """
        Provides
        """

        # FIXMEs:
        # (1) should be able to unify some of this code with BedDataProvider.process_data
        # (2) are optional number of parameters supported?

        # Build data to return. Payload format is:
        # [ <guid/offset>, <start>, <end>, <name>, <strand>, <thick_start>,
        #   <thick_end>, <blocks> ]
        #
        # First three entries are mandatory, others are optional.
        #
        no_detail = "no_detail" in kwargs
        rval = []
        message = None
        for count, line in enumerate(iterator):
            if count < start_val:
                continue
            if max_vals and count - start_val >= max_vals:
                message = self.error_max_vals % (max_vals, "features")
                break

            feature = line.split()

            # Feature initialization.
            payload: PAYLOAD_LIST_TYPE = [
                # GUID is just a hash of the line
                hash(line),
                # Add start, end.
                int(feature[1]),
                int(feature[2]),
            ]

            if no_detail:
                rval.append(payload)
                continue

            # Extend with additional data.
            payload.extend(
                [
                    # Add name, strand.
                    feature[3],
                    feature[5],
                    # Thick start, end are feature start, end for now.
                    int(feature[1]),
                    int(feature[2]),
                    # No blocks.
                    None,
                    # Filtering data: Score, signalValue, pValue, qValue.
                    float(feature[4]),
                    float(feature[6]),
                    float(feature[7]),
                    float(feature[8]),
                ]
            )

            rval.append(payload)

        return {"data": rval, "message": message}


class ENCODEPeakTabixDataProvider(TabixDataProvider, ENCODEPeakDataProvider):
    """
    Provides data from an ENCODEPeak dataset indexed via tabix.
    """

    def get_filters(self):
        """
        Returns filters for dataset.
        """
        # HACK: first 8 fields are for drawing, so start filter column index at 9.
        filter_col = 8
        filters = []
        filters.append(
            {"name": "Score", "type": "number", "index": filter_col, "tool_id": "Filter1", "tool_exp_name": "c6"}
        )
        filter_col += 1
        filters.append(
            {"name": "Signal Value", "type": "number", "index": filter_col, "tool_id": "Filter1", "tool_exp_name": "c7"}
        )
        filter_col += 1
        filters.append(
            {"name": "pValue", "type": "number", "index": filter_col, "tool_id": "Filter1", "tool_exp_name": "c8"}
        )
        filter_col += 1
        filters.append(
            {"name": "qValue", "type": "number", "index": filter_col, "tool_id": "Filter1", "tool_exp_name": "c9"}
        )
        return filters


#
# -- ChromatinInteraction data providers --
#


class ChromatinInteractionsDataProvider(GenomeDataProvider):
    def process_data(self, iterator, start_val=0, max_vals=None, **kwargs):
        """
        Provides
        """

        rval = []
        message = None
        for count, line in enumerate(iterator):
            if count < start_val:
                continue
            if max_vals and count - start_val >= max_vals:
                message = self.error_max_vals % (max_vals, "interactions")
                break

            feature = line.split()

            s1 = int(feature[1])
            e1 = int(feature[2])
            c = feature[3]
            s2 = int(feature[4])
            e2 = int(feature[5])
            v = float(feature[6])

            # Feature initialization.
            payload = [
                # GUID is just a hash of the line
                hash(line),
                # Add start1, end1, chr2, start2, end2, value.
                s1,
                e1,
                c,
                s2,
                e2,
                v,
            ]

            rval.append(payload)

        return {"data": rval, "message": message}

    def get_default_max_vals(self):
        return 100000


class ChromatinInteractionsTabixDataProvider(TabixDataProvider, ChromatinInteractionsDataProvider):
    def get_iterator(
        self, data_file, chrom, start=0, end=sys.maxsize, interchromosomal=False, **kwargs
    ) -> Iterator[str]:
        """ """
        # Modify start as needed to get earlier interactions with start region.
        span = int(end) - int(start)
        filter_start = max(0, int(start) - span - span / 2)

        def filter(iter):
            for line in iter:
                feature = line.split()
                s1 = int(feature[1])
                e1 = int(feature[2])
                c = feature[3]
                s2 = int(feature[4])
                e2 = int(feature[5])
                # Check for intrachromosal interactions.
                if ((s1 + s2) / 2 <= end) and ((e1 + e2) / 2 >= start) and (c == chrom):
                    yield line
                # Check for interchromosal interactions.
                if interchromosomal and c != chrom:
                    yield line

        return filter(TabixDataProvider.get_iterator(self, data_file, chrom, filter_start, end))


#
# -- Helper methods. --
#


def package_gff_feature(feature, no_detail=False, filter_cols=None) -> PAYLOAD_LIST_TYPE:
    """Package a GFF feature in an array for data providers."""
    filter_cols = filter_cols or []
    feature = convert_gff_coords_to_bed(feature)

    # No detail means only start, end.
    if no_detail:
        return [feature.start, feature.end]

    # Return full feature.
    payload = [
        feature.start,
        feature.end,
        feature.name(),
        feature.strand,
        # No notion of thick start, end in GFF, so make everything
        # thick.
        feature.start,
        feature.end,
    ]

    # HACK: ignore interval with name 'transcript' from feature.
    # Cufflinks puts this interval in each of its transcripts,
    # and they mess up trackster by covering the feature's blocks.
    # This interval will always be a feature's first interval,
    # and the GFF's third column is its feature name.
    feature_intervals = feature.intervals
    if feature.intervals[0].fields[2] == "transcript":
        feature_intervals = feature.intervals[1:]
    # Add blocks.
    block_sizes = [(interval.end - interval.start) for interval in feature_intervals]
    block_starts = [(interval.start - feature.start) for interval in feature_intervals]
    blocks = list(zip(block_sizes, block_starts))
    payload.append([(feature.start + block[1], feature.start + block[1] + block[0]) for block in blocks])

    # Add filter data to payload.
    for col in filter_cols:
        if col == "Score":
            if feature.score == "nan":
                payload.append(feature.score)
            else:
                try:
                    f = float(feature.score)
                    payload.append(f)
                except Exception:
                    payload.append(feature.score)
        elif col in feature.attributes:
            if feature.attributes[col] == "nan":
                payload.append(feature.attributes[col])
            else:
                try:
                    f = float(feature.attributes[col])
                    payload.append(f)
                except Exception:
                    payload.append(feature.attributes[col])
        else:
            # Dummy value.
            payload.append(0)
    return payload



import json
import logging
import os
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
)

from sqlalchemy import or_

from galaxy import (
    exceptions,
    util,
)
from galaxy.structured_app import StructuredApp
from galaxy.tool_shed.galaxy_install.metadata.installed_repository_metadata_manager import (
    InstalledRepositoryMetadataManager,
)
from galaxy.tool_shed.galaxy_install.repository_dependencies import repository_dependency_manager
from galaxy.tool_shed.galaxy_install.tools import (
    data_manager,
    tool_panel_manager,
)
from galaxy.tool_shed.tools.data_table_manager import ShedToolDataTableManager
from galaxy.tool_shed.util import (
    hg_util,
    repository_util,
    shed_util_common as suc,
    tool_util,
)
from galaxy.tool_util.deps import views
from galaxy.util.tool_shed import (
    common_util,
    encoding_util,
)

log = logging.getLogger(__name__)


class InstallRepositoryManager:
    app: StructuredApp
    tpm: tool_panel_manager.ToolPanelManager

    def __init__(self, app: StructuredApp, tpm: Optional[tool_panel_manager.ToolPanelManager] = None):
        self.app = app
        self.install_model = self.app.install_model
        self._view = views.DependencyResolversView(app)
        if tpm is None:
            self.tpm = tool_panel_manager.ToolPanelManager(self.app)
        else:
            self.tpm = tpm

    def _get_repository_components_for_installation(
        self, encoded_tsr_id, encoded_tsr_ids, repo_info_dicts, tool_panel_section_keys
    ):
        """
        The received encoded_tsr_ids, repo_info_dicts, and
        tool_panel_section_keys are 3 lists that contain associated elements
        at each location in the list.  This method will return the elements
        from repo_info_dicts and tool_panel_section_keys associated with the
        received encoded_tsr_id by determining its location in the received
        encoded_tsr_ids list.
        """
        for tsr_id, repo_info_dict, tool_panel_section_key in zip(
            encoded_tsr_ids, repo_info_dicts, tool_panel_section_keys
        ):
            if tsr_id == encoded_tsr_id:
                return repo_info_dict, tool_panel_section_key
        return None, None

    def __get_install_info_from_tool_shed(
        self, tool_shed_url: str, name: str, owner: str, changeset_revision: str
    ) -> Tuple[Dict[str, Any], List[Dict[str, Any]]]:
        params = dict(name=name, owner=owner, changeset_revision=changeset_revision)
        pathspec = ["api", "repositories", "get_repository_revision_install_info"]
        try:
            raw_text = util.url_get(
                tool_shed_url,
                auth=self.app.tool_shed_registry.url_auth(tool_shed_url),
                pathspec=pathspec,
                params=params,
            )
        except Exception:
            message = "Error attempting to retrieve installation information from tool shed "
            message += f"{tool_shed_url} for revision {changeset_revision} of repository {name} owned by {owner}"
            log.exception(message)
            raise exceptions.InternalServerError(message)
        if raw_text:
            # If successful, the response from get_repository_revision_install_info will be 3
            # dictionaries, a dictionary defining the Repository, a dictionary defining the
            # Repository revision (RepositoryMetadata), and a dictionary including the additional
            # information required to install the repository.
            items = json.loads(util.unicodify(raw_text))
            repository_revision_dict = items[1]
            repo_info_dict = items[2]
        else:
            message = (
                "Unable to retrieve installation information from tool shed %s for revision %s of repository %s owned by %s"
                % (str(tool_shed_url), str(changeset_revision), str(name), str(owner))
            )
            log.warning(message)
            raise exceptions.InternalServerError(message)
        # Make sure the tool shed returned everything we need for installing the repository.
        if not repository_revision_dict or not repo_info_dict:
            invalid_parameter_message = "No information is available for the requested repository revision.\n"
            invalid_parameter_message += "One or more of the following parameter values is likely invalid:\n"
            invalid_parameter_message += f"tool_shed_url: {tool_shed_url}\n"
            invalid_parameter_message += f"name: {name}\n"
            invalid_parameter_message += f"owner: {owner}\n"
            invalid_parameter_message += f"changeset_revision: {changeset_revision}\n"
            raise exceptions.RequestParameterInvalidException(invalid_parameter_message)
        repo_info_dicts = [repo_info_dict]
        return repository_revision_dict, repo_info_dicts

    def __handle_repository_contents(
        self,
        tool_shed_repository,
        tool_path,
        repository_clone_url,
        relative_install_dir,
        tool_shed=None,
        tool_section=None,
        shed_tool_conf=None,
        reinstalling=False,
        tool_panel_section_mapping=None,
    ) -> None:
        """
        Generate the metadata for the installed tool shed repository, among other things.
        This method is called when an administrator is installing a new repository or
        reinstalling an uninstalled repository.
        """
        tool_panel_section_mapping = tool_panel_section_mapping or {}
        shed_config_dict = self.app.toolbox.get_shed_config_dict_by_filename(shed_tool_conf)
        stdtm = ShedToolDataTableManager(self.app)
        irmm = InstalledRepositoryMetadataManager(
            app=self.app,
            tpm=self.tpm,
            repository=tool_shed_repository,
            changeset_revision=tool_shed_repository.changeset_revision,
            repository_clone_url=repository_clone_url,
            shed_config_dict=shed_config_dict,
            relative_install_dir=relative_install_dir,
            repository_files_dir=None,
            resetting_all_metadata_on_repository=False,
            updating_installed_repository=False,
            persist=True,
        )
        irmm.generate_metadata_for_changeset_revision()
        irmm_metadata_dict = irmm.get_metadata_dict()
        tool_shed_repository.metadata_ = irmm_metadata_dict
        # Update the tool_shed_repository.tool_shed_status column in the database.
        tool_shed_status_dict = repository_util.get_tool_shed_status_for_installed_repository(
            self.app, tool_shed_repository
        )
        if tool_shed_status_dict:
            tool_shed_repository.tool_shed_status = tool_shed_status_dict
        self.install_model.context.add(tool_shed_repository)
        self.install_model.context.flush()
        if "sample_files" in irmm_metadata_dict:
            sample_files = irmm_metadata_dict.get("sample_files", [])
            tool_index_sample_files = stdtm.get_tool_index_sample_files(sample_files)
            tool_data_table_conf_filename, tool_data_table_elems = stdtm.install_tool_data_tables(
                tool_shed_repository, tool_index_sample_files
            )
            if tool_data_table_elems:
                self.app.tool_data_tables.add_new_entries_from_config_file(
                    tool_data_table_conf_filename, None, self.app.config.shed_tool_data_table_config, persist=True
                )
        if "tools" in irmm_metadata_dict:
            tool_panel_dict = self.tpm.generate_tool_panel_dict_for_new_install(
                irmm_metadata_dict["tools"], tool_section
            )
            sample_files = irmm_metadata_dict.get("sample_files", [])
            tool_index_sample_files = stdtm.get_tool_index_sample_files(sample_files)
            tool_data_path = self.app.config.tool_data_path
            tool_util.copy_sample_files(tool_data_path, tool_index_sample_files, tool_path=tool_path)
            sample_files_copied = [str(s) for s in tool_index_sample_files]
            repository_tools_tups = irmm.get_repository_tools_tups()
            if repository_tools_tups:
                # Handle missing data table entries for tool parameters that are dynamically generated select lists.
                repository_tools_tups = stdtm.handle_missing_data_table_entry(
                    relative_install_dir, tool_path, repository_tools_tups
                )
                # Handle missing index files for tool parameters that are dynamically generated select lists.
                repository_tools_tups, sample_files_copied = tool_util.handle_missing_index_file(
                    self.app, tool_path, sample_files, repository_tools_tups, sample_files_copied
                )
                # Copy remaining sample files included in the repository to the ~/tool-data directory of the
                # local Galaxy instance.
                tool_util.copy_sample_files(
                    tool_data_path, sample_files, tool_path=tool_path, sample_files_copied=sample_files_copied
                )
                self.tpm.add_to_tool_panel(
                    repository_name=tool_shed_repository.name,
                    repository_clone_url=repository_clone_url,
                    changeset_revision=tool_shed_repository.installed_changeset_revision,
                    repository_tools_tups=repository_tools_tups,
                    owner=tool_shed_repository.owner,
                    shed_tool_conf=shed_tool_conf,
                    tool_panel_dict=tool_panel_dict,
                    new_install=True,
                    tool_panel_section_mapping=tool_panel_section_mapping,
                )
        if "data_manager" in irmm_metadata_dict:
            dmh = data_manager.DataManagerHandler(self.app)
            assert shed_config_dict
            dmh.install_data_managers(
                self.app.config.shed_data_manager_config_file,
                irmm_metadata_dict,
                shed_config_dict,
                relative_install_dir,
                tool_shed_repository,
                repository_tools_tups,
            )
        if "datatypes" in irmm_metadata_dict:
            log.warning("Ignoring tool shed datatypes... these have been deprecated.")

    def handle_tool_shed_repositories(self, installation_dict):
        # The following installation_dict entries are all required.
        install_repository_dependencies = installation_dict["install_repository_dependencies"]
        new_tool_panel_section_label = installation_dict["new_tool_panel_section_label"]
        no_changes_checked = installation_dict["no_changes_checked"]
        repo_info_dicts = installation_dict["repo_info_dicts"]
        tool_panel_section_id = installation_dict["tool_panel_section_id"]
        tool_path = installation_dict["tool_path"]
        tool_shed_url = installation_dict["tool_shed_url"]
        rdim = repository_dependency_manager.RepositoryDependencyInstallManager(self.app)
        (
            created_or_updated_tool_shed_repositories,
            tool_panel_section_keys,
            repo_info_dicts,
            filtered_repo_info_dicts,
        ) = rdim.create_repository_dependency_objects(
            tool_path=tool_path,
            tool_shed_url=tool_shed_url,
            repo_info_dicts=repo_info_dicts,
            install_repository_dependencies=install_repository_dependencies,
            no_changes_checked=no_changes_checked,
            tool_panel_section_id=tool_panel_section_id,
            new_tool_panel_section_label=new_tool_panel_section_label,
        )
        return (
            created_or_updated_tool_shed_repositories,
            tool_panel_section_keys,
            repo_info_dicts,
            filtered_repo_info_dicts,
        )

    def initiate_repository_installation(self, installation_dict):
        # The following installation_dict entries are all required.
        created_or_updated_tool_shed_repositories = installation_dict["created_or_updated_tool_shed_repositories"]
        filtered_repo_info_dicts = installation_dict["filtered_repo_info_dicts"]
        has_repository_dependencies = installation_dict["has_repository_dependencies"]
        includes_tool_dependencies = installation_dict["includes_tool_dependencies"]
        includes_tools = installation_dict["includes_tools"]
        includes_tools_for_display_in_tool_panel = installation_dict["includes_tools_for_display_in_tool_panel"]
        install_repository_dependencies = installation_dict["install_repository_dependencies"]
        install_resolver_dependencies = installation_dict["install_resolver_dependencies"]
        install_tool_dependencies = installation_dict["install_tool_dependencies"]
        message = installation_dict["message"]
        new_tool_panel_section_label = installation_dict["new_tool_panel_section_label"]
        shed_tool_conf = installation_dict["shed_tool_conf"]
        status = installation_dict["status"]
        tool_panel_section_id = installation_dict["tool_panel_section_id"]
        tool_panel_section_keys = installation_dict["tool_panel_section_keys"]
        tool_panel_section_mapping = installation_dict.get("tool_panel_section_mapping", {})
        tool_path = installation_dict["tool_path"]
        tool_shed_url = installation_dict["tool_shed_url"]
        # Handle contained tools.
        if includes_tools_for_display_in_tool_panel and (new_tool_panel_section_label or tool_panel_section_id):
            self.tpm.handle_tool_panel_section(
                self.app.toolbox,
                tool_panel_section_id=tool_panel_section_id,
                new_tool_panel_section_label=new_tool_panel_section_label,
            )
        if includes_tools_for_display_in_tool_panel and (tool_panel_section_mapping is not None):
            for tool_guid in tool_panel_section_mapping:
                if tool_panel_section_mapping[tool_guid]["action"] == "create":
                    new_tool_panel_section_name = tool_panel_section_mapping[tool_guid]["tool_panel_section"]
                    log.debug(f'Creating tool panel section "{new_tool_panel_section_name}" for tool {tool_guid}')
                    self.tpm.handle_tool_panel_section(
                        self.app.toolbox, None, tool_panel_section_mapping[tool_guid]["tool_panel_section"]
                    )
        encoded_repository_ids = [
            self.app.security.encode_id(tsr.id) for tsr in created_or_updated_tool_shed_repositories
        ]
        new_kwd = dict(
            includes_tools=includes_tools,
            includes_tools_for_display_in_tool_panel=includes_tools_for_display_in_tool_panel,
            has_repository_dependencies=has_repository_dependencies,
            install_repository_dependencies=install_repository_dependencies,
            includes_tool_dependencies=includes_tool_dependencies,
            install_resolver_dependencies=install_resolver_dependencies,
            install_tool_dependencies=install_tool_dependencies,
            message=message,
            repo_info_dicts=filtered_repo_info_dicts,
            shed_tool_conf=shed_tool_conf,
            status=status,
            tool_path=tool_path,
            tool_panel_section_keys=tool_panel_section_keys,
            tool_shed_repository_ids=encoded_repository_ids,
            tool_shed_url=tool_shed_url,
        )
        encoded_kwd = encoding_util.tool_shed_encode(new_kwd)
        tsr_ids = [r.id for r in created_or_updated_tool_shed_repositories]
        tool_shed_repositories = []
        for tsr_id in tsr_ids:
            tsr = self.install_model.context.query(self.install_model.ToolShedRepository).get(tsr_id)
            tool_shed_repositories.append(tsr)
        clause_list = []
        for tsr_id in tsr_ids:
            clause_list.append(self.install_model.ToolShedRepository.table.c.id == tsr_id)
        query = self.install_model.context.query(self.install_model.ToolShedRepository).filter(or_(*clause_list))
        return encoded_kwd, query, tool_shed_repositories, encoded_repository_ids

    def install(
        self, tool_shed_url: str, name: str, owner: str, changeset_revision: str, install_options: Dict[str, Any]
    ):
        # Get all of the information necessary for installing the repository from the specified tool shed.
        repository_revision_dict, repo_info_dicts = self.__get_install_info_from_tool_shed(
            tool_shed_url, name, owner, changeset_revision
        )
        if changeset_revision != repository_revision_dict["changeset_revision"]:
            # Demanded installation of a non-installable revision. Stop here if repository already installed.
            repo = repository_util.get_installed_repository(
                app=self.app,
                tool_shed=tool_shed_url,
                name=name,
                owner=owner,
                changeset_revision=repository_revision_dict["changeset_revision"],
            )
            if repo and repo.is_installed:
                # Repo installed. Returning empty list indicated repo already installed.
                return []
        installed_tool_shed_repositories = self.__initiate_and_install_repositories(
            tool_shed_url, repository_revision_dict, repo_info_dicts, install_options
        )
        return installed_tool_shed_repositories

    def __initiate_and_install_repositories(
        self,
        tool_shed_url: str,
        repository_revision_dict: Dict[str, Any],
        repo_info_dicts: List[Dict[str, Any]],
        install_options: Dict[str, Any],
    ):
        try:
            has_repository_dependencies = repository_revision_dict["has_repository_dependencies"]
        except KeyError:
            raise exceptions.InternalServerError(
                "Tool shed response missing required parameter 'has_repository_dependencies'."
            )
        try:
            includes_tools = repository_revision_dict["includes_tools"]
        except KeyError:
            raise exceptions.InternalServerError("Tool shed response missing required parameter 'includes_tools'.")
        try:
            includes_tool_dependencies = repository_revision_dict["includes_tool_dependencies"]
        except KeyError:
            raise exceptions.InternalServerError(
                "Tool shed response missing required parameter 'includes_tool_dependencies'."
            )
        try:
            includes_tools_for_display_in_tool_panel = repository_revision_dict[
                "includes_tools_for_display_in_tool_panel"
            ]
        except KeyError:
            raise exceptions.InternalServerError(
                "Tool shed response missing required parameter 'includes_tools_for_display_in_tool_panel'."
            )
        # Get the information about the Galaxy components (e.g., tool panel section, tool config file, etc) that will contain the repository information.
        install_repository_dependencies = install_options.get("install_repository_dependencies", False)
        install_resolver_dependencies = install_options.get("install_resolver_dependencies", False)
        install_tool_dependencies = install_options.get("install_tool_dependencies", False)
        new_tool_panel_section_label = install_options.get("new_tool_panel_section_label", "")
        tool_panel_section_mapping = install_options.get("tool_panel_section_mapping", {})
        shed_tool_conf = install_options.get("shed_tool_conf")
        if install_tool_dependencies and self.app.tool_dependency_dir is None:
            raise exceptions.ConfigDoesNotAllowException(
                "Tool dependency installation is disabled in your configuration files."
            )
        if shed_tool_conf:
            # Get the tool_path setting.
            shed_config_dict = self.tpm.get_shed_tool_conf_dict(shed_tool_conf)
        else:
            shed_config_dict = self.app.toolbox.default_shed_tool_conf_dict()
        shed_tool_conf = shed_config_dict["config_filename"]
        tool_path = shed_config_dict["tool_path"]
        tool_panel_section_id = self.app.toolbox.find_section_id(install_options.get("tool_panel_section_id", ""))
        # Build the dictionary of information necessary for creating tool_shed_repository database records
        # for each repository being installed.
        installation_dict = dict(
            install_repository_dependencies=install_repository_dependencies,
            new_tool_panel_section_label=new_tool_panel_section_label,
            tool_panel_section_mapping=tool_panel_section_mapping,
            no_changes_checked=False,
            repo_info_dicts=repo_info_dicts,
            tool_panel_section_id=tool_panel_section_id,
            tool_path=tool_path,
            tool_shed_url=tool_shed_url,
        )
        # Create the tool_shed_repository database records and gather additional information for repository installation.
        (
            created_or_updated_tool_shed_repositories,
            tool_panel_section_keys,
            repo_info_dicts,
            filtered_repo_info_dicts,
        ) = self.handle_tool_shed_repositories(installation_dict)
        if created_or_updated_tool_shed_repositories:
            # Build the dictionary of information necessary for installing the repositories.
            installation_dict = dict(
                created_or_updated_tool_shed_repositories=created_or_updated_tool_shed_repositories,
                filtered_repo_info_dicts=filtered_repo_info_dicts,
                has_repository_dependencies=has_repository_dependencies,
                includes_tool_dependencies=includes_tool_dependencies,
                includes_tools=includes_tools,
                includes_tools_for_display_in_tool_panel=includes_tools_for_display_in_tool_panel,
                install_repository_dependencies=install_repository_dependencies,
                install_resolver_dependencies=install_resolver_dependencies,
                install_tool_dependencies=install_tool_dependencies,
                message="",
                new_tool_panel_section_label=new_tool_panel_section_label,
                shed_tool_conf=shed_tool_conf,
                status="done",
                tool_panel_section_id=tool_panel_section_id,
                tool_panel_section_keys=tool_panel_section_keys,
                tool_panel_section_mapping=tool_panel_section_mapping,
                tool_path=tool_path,
                tool_shed_url=tool_shed_url,
            )
            # Prepare the repositories for installation.  Even though this
            # method receives a single combination of tool_shed_url, name,
            # owner and changeset_revision, there may be multiple repositories
            # for installation at this point because repository dependencies
            # may have added additional repositories for installation along
            # with the single specified repository.
            encoded_kwd, query, tool_shed_repositories, encoded_repository_ids = self.initiate_repository_installation(
                installation_dict
            )
            # Some repositories may have repository dependencies that are
            # required to be installed before the dependent repository, so
            # we'll order the list of tsr_ids to ensure all repositories
            # install in the required order.
            tsr_ids = [
                self.app.security.encode_id(tool_shed_repository.id) for tool_shed_repository in tool_shed_repositories
            ]

            decoded_kwd = dict(
                shed_tool_conf=shed_tool_conf,
                tool_path=tool_path,
                tool_panel_section_keys=tool_panel_section_keys,
                repo_info_dicts=filtered_repo_info_dicts,
                install_resolver_dependencies=install_resolver_dependencies,
                install_tool_dependencies=install_tool_dependencies,
                tool_panel_section_mapping=tool_panel_section_mapping,
            )
            return self.install_repositories(tsr_ids, decoded_kwd, reinstalling=False, install_options=install_options)

    def install_repositories(self, tsr_ids, decoded_kwd, reinstalling, install_options=None):
        shed_tool_conf = decoded_kwd.get("shed_tool_conf", "")
        tool_path = decoded_kwd["tool_path"]
        tool_panel_section_keys = util.listify(decoded_kwd["tool_panel_section_keys"])
        tool_panel_section_mapping = decoded_kwd.get("tool_panel_section_mapping", {})
        repo_info_dicts = util.listify(decoded_kwd["repo_info_dicts"])
        install_resolver_dependencies = decoded_kwd["install_resolver_dependencies"]
        install_tool_dependencies = decoded_kwd["install_tool_dependencies"]
        filtered_repo_info_dicts = []
        filtered_tool_panel_section_keys = []
        repositories_for_installation = []
        # Some repositories may have repository dependencies that are required to be installed before the
        # dependent repository, so we'll order the list of tsr_ids to ensure all repositories install in the
        # required order.
        (
            ordered_tsr_ids,
            ordered_repo_info_dicts,
            ordered_tool_panel_section_keys,
        ) = self.order_components_for_installation(
            tsr_ids, repo_info_dicts, tool_panel_section_keys=tool_panel_section_keys
        )
        for tsr_id in ordered_tsr_ids:
            repository = self.install_model.context.query(self.install_model.ToolShedRepository).get(
                self.app.security.decode_id(tsr_id)
            )
            if repository.status in [
                self.install_model.ToolShedRepository.installation_status.NEW,
                self.install_model.ToolShedRepository.installation_status.UNINSTALLED,
            ]:
                repositories_for_installation.append(repository)
                repo_info_dict, tool_panel_section_key = self._get_repository_components_for_installation(
                    tsr_id, ordered_tsr_ids, ordered_repo_info_dicts, ordered_tool_panel_section_keys
                )
                filtered_repo_info_dicts.append(repo_info_dict)
                filtered_tool_panel_section_keys.append(tool_panel_section_key)

        installed_tool_shed_repositories = []
        if repositories_for_installation:
            for tool_shed_repository, repo_info_dict, tool_panel_section_key in zip(
                repositories_for_installation, filtered_repo_info_dicts, filtered_tool_panel_section_keys
            ):
                pre_install_state = tool_shed_repository.status
                try:
                    self.install_tool_shed_repository(
                        tool_shed_repository,
                        repo_info_dict=repo_info_dict,
                        tool_panel_section_key=tool_panel_section_key,
                        shed_tool_conf=shed_tool_conf,
                        tool_path=tool_path,
                        install_resolver_dependencies=install_resolver_dependencies,
                        install_tool_dependencies=install_tool_dependencies,
                        reinstalling=reinstalling,
                        tool_panel_section_mapping=tool_panel_section_mapping,
                    )
                except Exception as e:
                    log.exception("Error installing repository '%s'", tool_shed_repository.name)
                    if pre_install_state != self.install_model.ToolShedRepository.states.OK:
                        # If repository was in OK state previously and e.g. an update failed don't set the state to ERROR.
                        # For every other state do update the state to error and reset files on disk,
                        # so that another attempt can be made
                        repository_util.set_repository_attributes(
                            self.app,
                            tool_shed_repository,
                            status=self.install_model.ToolShedRepository.installation_status.ERROR,
                            error_message=util.unicodify(e),
                            deleted=False,
                            uninstalled=False,
                            remove_from_disk=True,
                        )
                installed_tool_shed_repositories.append(tool_shed_repository)
        else:
            raise RepositoriesInstalledException()
        return installed_tool_shed_repositories

    def install_tool_shed_repository(
        self,
        tool_shed_repository,
        repo_info_dict,
        tool_panel_section_key,
        shed_tool_conf,
        tool_path,
        install_resolver_dependencies,
        install_tool_dependencies,
        reinstalling=False,
        tool_panel_section_mapping=None,
        install_options=None,
    ):
        tool_panel_section_mapping = tool_panel_section_mapping or {}
        self.app.install_model.context.flush()
        if tool_panel_section_key:
            _, tool_section = self.app.toolbox.get_section(tool_panel_section_key)
            if tool_section is None:
                log.debug(
                    'Invalid tool_panel_section_key "%s" specified.  Tools will be loaded outside of sections in the tool panel.',
                    str(tool_panel_section_key),
                )
        else:
            tool_section = None
        if isinstance(repo_info_dict, str):
            repo_info_dict = encoding_util.tool_shed_decode(repo_info_dict)
        repo_info_tuple = repo_info_dict[tool_shed_repository.name]
        (
            description,
            repository_clone_url,
            changeset_revision,
            ctx_rev,
            repository_owner,
            repository_dependencies,
            tool_dependencies,
        ) = repo_info_tuple
        if changeset_revision != tool_shed_repository.changeset_revision:
            # This is an update
            tool_shed_url = common_util.get_tool_shed_url_from_tool_shed_registry(
                self.app, tool_shed_repository.tool_shed
            )
            return self.update_tool_shed_repository(
                tool_shed_repository, tool_shed_url, ctx_rev, changeset_revision, install_options=install_options
            )
        # Clone the repository to the configured location.
        self.update_tool_shed_repository_status(
            tool_shed_repository, self.install_model.ToolShedRepository.installation_status.CLONING
        )
        relative_clone_dir = repository_util.generate_tool_shed_repository_install_dir(
            repository_clone_url, tool_shed_repository.installed_changeset_revision
        )
        relative_install_dir = os.path.join(relative_clone_dir, tool_shed_repository.name)
        install_dir = os.path.abspath(os.path.join(tool_path, relative_install_dir))
        log.info(
            "Cloning repository '%s' at %s:%s", repository_clone_url, ctx_rev, tool_shed_repository.changeset_revision
        )
        if os.path.exists(install_dir):
            # May exist from a previous failed install attempt, just try updating instead of cloning.
            hg_util.pull_repository(install_dir, repository_clone_url, ctx_rev)
            hg_util.update_repository(install_dir, ctx_rev)
        else:
            _, error_message = hg_util.clone_repository(repository_clone_url, install_dir, ctx_rev)
            if error_message:
                raise Exception(error_message)
        if reinstalling:
            # Since we're reinstalling the repository we need to find the latest changeset revision to
            # which it can be updated.
            update_to_changeset = self.app.update_repository_manager.get_update_to_changeset_revision_and_ctx_rev(
                tool_shed_repository
            )
            current_changeset_revision = update_to_changeset.changeset_revision
            current_ctx_rev = update_to_changeset.ctx_rev
            if current_ctx_rev != ctx_rev:
                repo_path = os.path.abspath(install_dir)
                hg_util.pull_repository(repo_path, repository_clone_url, current_changeset_revision)
                hg_util.update_repository(repo_path, ctx_rev=current_ctx_rev)
        self.__handle_repository_contents(
            tool_shed_repository=tool_shed_repository,
            tool_path=tool_path,
            repository_clone_url=repository_clone_url,
            relative_install_dir=relative_install_dir,
            tool_shed=tool_shed_repository.tool_shed,
            tool_section=tool_section,
            shed_tool_conf=shed_tool_conf,
            reinstalling=reinstalling,
            tool_panel_section_mapping=tool_panel_section_mapping,
        )
        metadata = tool_shed_repository.metadata_
        if "tools" in metadata and install_resolver_dependencies:
            self.update_tool_shed_repository_status(
                tool_shed_repository,
                self.install_model.ToolShedRepository.installation_status.INSTALLING_TOOL_DEPENDENCIES,
            )
            new_tools = [self.app.toolbox._tools_by_id.get(tool_d["guid"], None) for tool_d in metadata["tools"]]
            new_requirements = {tool.requirements.packages for tool in new_tools if tool}
            [self._view.install_dependencies(r) for r in new_requirements]
            dependency_manager = self.app.toolbox.dependency_manager
            if dependency_manager.cached:
                [dependency_manager.build_cache(r) for r in new_requirements]

        self.update_tool_shed_repository_status(
            tool_shed_repository, self.install_model.ToolShedRepository.installation_status.INSTALLED
        )

    def update_tool_shed_repository(
        self,
        repository,
        tool_shed_url,
        latest_ctx_rev,
        latest_changeset_revision,
        install_new_dependencies=True,
        install_options=None,
    ):
        install_options = install_options or {}
        original_metadata_dict = repository.metadata_
        original_repository_dependencies_dict = original_metadata_dict.get("repository_dependencies", {})
        original_repository_dependencies = original_repository_dependencies_dict.get("repository_dependencies", [])
        shed_tool_conf, tool_path, relative_install_dir = suc.get_tool_panel_config_tool_path_install_dir(
            self.app, repository
        )
        if tool_path:
            repo_files_dir = os.path.abspath(os.path.join(tool_path, relative_install_dir, repository.name))
        else:
            repo_files_dir = os.path.abspath(os.path.join(relative_install_dir, repository.name))
        repository_clone_url = os.path.join(tool_shed_url, "repos", repository.owner, repository.name)
        # Set a status, even though we're probably not cloning.
        self.update_tool_shed_repository_status(repository, status=repository.installation_status.CLONING)
        log.info("Updating repository '%s' to %s:%s", repository.name, latest_ctx_rev, latest_changeset_revision)
        if not os.path.exists(repo_files_dir):
            log.debug(
                "Repository directory '%s' does not exist, cloning repository instead of updating repository",
                repo_files_dir,
            )
            hg_util.clone_repository(
                repository_clone_url=repository_clone_url, repository_file_dir=repo_files_dir, ctx_rev=latest_ctx_rev
            )
        hg_util.pull_repository(repo_files_dir, repository_clone_url, latest_ctx_rev)
        hg_util.update_repository(repo_files_dir, latest_ctx_rev)
        # Remove old Data Manager entries
        if repository.includes_data_managers:
            dmh = data_manager.DataManagerHandler(self.app)
            dmh.remove_from_data_manager(repository)
        # Update the repository metadata.
        tpm = tool_panel_manager.ToolPanelManager(self.app)
        irmm = InstalledRepositoryMetadataManager(
            app=self.app,
            tpm=tpm,
            repository=repository,
            changeset_revision=latest_changeset_revision,
            repository_clone_url=repository_clone_url,
            shed_config_dict=repository.get_shed_config_dict(self.app),
            relative_install_dir=relative_install_dir,
            repository_files_dir=None,
            resetting_all_metadata_on_repository=False,
            updating_installed_repository=True,
            persist=True,
        )
        irmm.generate_metadata_for_changeset_revision()
        irmm_metadata_dict = irmm.get_metadata_dict()
        self.update_tool_shed_repository_status(repository, status=repository.installation_status.INSTALLED)
        if "tools" in irmm_metadata_dict:
            tool_panel_dict = irmm_metadata_dict.get("tool_panel_section", None)
            if tool_panel_dict is None:
                tool_panel_dict = tpm.generate_tool_panel_dict_from_shed_tool_conf_entries(repository)
            repository_tools_tups = irmm.get_repository_tools_tups()
            tpm.add_to_tool_panel(
                repository_name=str(repository.name),
                repository_clone_url=repository_clone_url,
                changeset_revision=str(repository.installed_changeset_revision),
                repository_tools_tups=repository_tools_tups,
                owner=str(repository.owner),
                shed_tool_conf=shed_tool_conf,
                tool_panel_dict=tool_panel_dict,
                new_install=False,
            )
            # Add new Data Manager entries
            if "data_manager" in irmm_metadata_dict:
                dmh = data_manager.DataManagerHandler(self.app)
                dmh.install_data_managers(
                    self.app.config.shed_data_manager_config_file,
                    irmm_metadata_dict,
                    repository.get_shed_config_dict(self.app),
                    os.path.join(relative_install_dir, repository.name),
                    repository,
                    repository_tools_tups,
                )
        if "repository_dependencies" in irmm_metadata_dict or "tool_dependencies" in irmm_metadata_dict:
            new_repository_dependencies_dict = irmm_metadata_dict.get("repository_dependencies", {})
            new_repository_dependencies = new_repository_dependencies_dict.get("repository_dependencies", [])
            if new_repository_dependencies:
                # [[http://localhost:9009', package_picard_1_56_0', devteam', 910b0b056666', False', False']]
                if new_repository_dependencies == original_repository_dependencies:
                    for new_repository_tup in new_repository_dependencies:
                        # Make sure all dependencies are installed.
                        # TODO: Repository dependencies that are not installed should be displayed to the user,
                        # giving them the option to install them or not. This is the same behavior as when initially
                        # installing and when re-installing.
                        (
                            new_tool_shed,
                            new_name,
                            new_owner,
                            new_changeset_revision,
                            new_pir,
                            new_oicct,
                        ) = common_util.parse_repository_dependency_tuple(new_repository_tup)
                        # Mock up a repo_info_tupe that has the information needed to see if the repository dependency
                        # was previously installed.
                        repo_info_tuple = ("", new_tool_shed, new_changeset_revision, "", new_owner, [], [])
                        # Since the value of new_changeset_revision came from a repository dependency
                        # definition, it may occur earlier in the Tool Shed's repository changelog than
                        # the Galaxy tool_shed_repository.installed_changeset_revision record value, so
                        # we set from_tip to True to make sure we get the entire set of changeset revisions
                        # from the Tool Shed.
                        (
                            new_repository_db_record,
                            installed_changeset_revision,
                        ) = repository_util.repository_was_previously_installed(
                            self.app, tool_shed_url, new_name, repo_info_tuple, from_tip=True
                        )
                        if (
                            new_repository_db_record
                            and new_repository_db_record.status
                            in [
                                self.install_model.ToolShedRepository.installation_status.ERROR,
                                self.install_model.ToolShedRepository.installation_status.NEW,
                                self.install_model.ToolShedRepository.installation_status.UNINSTALLED,
                            ]
                        ) or not new_repository_db_record:
                            log.debug(
                                "Update to %s contains new repository dependency %s/%s",
                                repository.name,
                                new_owner,
                                new_name,
                            )
                            if not install_new_dependencies:
                                return ("repository", irmm_metadata_dict)
                            else:
                                self.install(
                                    tool_shed_url, new_name, new_owner, new_changeset_revision, install_options
                                )
        # Updates received did not include any newly defined repository dependencies or newly defined
        # tool dependencies that need to be installed.
        repository = self.app.update_repository_manager.update_repository_record(
            repository=repository,
            updated_metadata_dict=irmm_metadata_dict,
            updated_changeset_revision=latest_changeset_revision,
            updated_ctx_rev=latest_ctx_rev,
        )
        return (None, None)

    def order_components_for_installation(self, tsr_ids, repo_info_dicts, tool_panel_section_keys):
        """
        Some repositories may have repository dependencies that are required to be installed
        before the dependent repository.  This method will inspect the list of repositories
        about to be installed and make sure to order them appropriately.  For each repository
        about to be installed, if required repositories are not contained in the list of repositories
        about to be installed, then they are not considered.  Repository dependency definitions
        that contain circular dependencies should not result in an infinite loop, but obviously
        prior installation will not be handled for one or more of the repositories that require
        prior installation.
        """
        ordered_tsr_ids = []
        ordered_repo_info_dicts = []
        ordered_tool_panel_section_keys = []
        # Create a dictionary whose keys are the received tsr_ids and whose values are a list of
        # tsr_ids, each of which is contained in the received list of tsr_ids and whose associated
        # repository must be installed prior to the repository associated with the tsr_id key.
        prior_install_required_dict = repository_util.get_prior_import_or_install_required_dict(
            self.app, tsr_ids, repo_info_dicts
        )
        processed_tsr_ids = []
        while len(processed_tsr_ids) != len(prior_install_required_dict.keys()):
            tsr_id = suc.get_next_prior_import_or_install_required_dict_entry(
                prior_install_required_dict, processed_tsr_ids
            )
            processed_tsr_ids.append(tsr_id)
            # Create the ordered_tsr_ids, the ordered_repo_info_dicts and the ordered_tool_panel_section_keys lists.
            if tsr_id not in ordered_tsr_ids:
                prior_install_required_ids = prior_install_required_dict[tsr_id]
                for prior_install_required_id in prior_install_required_ids:
                    if prior_install_required_id not in ordered_tsr_ids:
                        # Install the associated repository dependency first.
                        (
                            prior_repo_info_dict,
                            prior_tool_panel_section_key,
                        ) = self._get_repository_components_for_installation(
                            prior_install_required_id,
                            tsr_ids,
                            repo_info_dicts,
                            tool_panel_section_keys=tool_panel_section_keys,
                        )
                        ordered_tsr_ids.append(prior_install_required_id)
                        ordered_repo_info_dicts.append(prior_repo_info_dict)
                        ordered_tool_panel_section_keys.append(prior_tool_panel_section_key)
                repo_info_dict, tool_panel_section_key = self._get_repository_components_for_installation(
                    tsr_id, tsr_ids, repo_info_dicts, tool_panel_section_keys=tool_panel_section_keys
                )
                if tsr_id not in ordered_tsr_ids:
                    ordered_tsr_ids.append(tsr_id)
                    ordered_repo_info_dicts.append(repo_info_dict)
                    ordered_tool_panel_section_keys.append(tool_panel_section_key)
        return ordered_tsr_ids, ordered_repo_info_dicts, ordered_tool_panel_section_keys

    def update_tool_shed_repository_status(self, tool_shed_repository, status, error_message=None):
        """
        Update the status of a tool shed repository in the process of being installed into Galaxy.
        """
        tool_shed_repository.status = status
        tool_shed_repository.error_message = error_message
        self.install_model.context.add(tool_shed_repository)
        self.install_model.context.flush()


class RepositoriesInstalledException(exceptions.RequestParameterInvalidException):
    def __init__(self):
        super().__init__("All repositories that you are attempting to install have been previously installed.")



"""Pre-configured iSEE options for selection by user.

Currently user configuration is NOT IMPLEMENTED - the code is here but option
is hidden from user in the tool form. Without ``custom`` selection, this simply
returns a DEFAULT iSEE configuration (defined at the bottom of this file).

These are preconfigured iSEE parameters that can be chosen by the user
as multiple choice (checkbox/select) input fields. Each parameter
(e.g. "initial") has a list of options whose indices should match the value of
the input field in the tool XML.

In all likelihood, the majority of options in
iSEE don't make sense without first being able to sniff the data and so are not
feasible to expose in the Galaxy tool form. Currently you will see that only
plot type and width can be exposed.
"""


def app():
    """Render R code to create iSEE app from user input."""
    return DEFAULT


def render_plots(call, plots):
    """Render plot calls from user input."""
    if not plots:
        return call
    plot_calls_list = [
        get_render_func(plot)(
            # user plot params as kwargs here
        )
        for plot in plots
    ]
    plot_calls = ",\ninitial=c(\n" + ",\n".join(plot_calls_list) + ")"
    return call + plot_calls


def get_render_func(plot):
    """Return the appropriate function to render plot."""
    # This is probably broken and unused
    return OPTIONS["plots"][plot["plot_types"]["plot_type"].value]  # type: ignore[index]


def reduced_dimension_plot(pw="6L"):
    """Render a ReducedDimensionPlot object call."""
    return f"""ReducedDimensionPlot(
        PanelWidth={pw})"""


def feature_assay_plot(pw="6L"):
    """Render a FeatureAssayPlot object call."""
    return f"""FeatureAssayPlot(
        PanelWidth={pw})"""


def row_data_table(pw="12L"):
    """Render a RowDataTable object call."""
    return f"RowDataTable(PanelWidth={pw})"


def column_data_plot(pw="6L"):
    """Render a ColumnDataPlot object call."""
    return f"ColumnDataPlot(PanelWidth={pw})"


OPTIONS = {
    "plots": {
        "reduced_dimension_plot": reduced_dimension_plot,
        "feature_assay_plot": feature_assay_plot,
        "row_data_table": row_data_table,
        "column_data_plot": column_data_plot,
    },
    "colormaps": {},
    "extra": {},
}


DEFAULT = """
sce <- registerAppOptions(sce, color.maxlevels=40)

categorical_color_fun <- function(n){
  if (n <= 37) {
    # Less than 37 colours, use something from colour brewer
    # (joining a bunch of palettes, best colours up front)
    multiset <-  c(
        RColorBrewer::brewer.pal(9, "Set1"),
        RColorBrewer::brewer.pal(8, "Set2"),
        RColorBrewer::brewer.pal(12, "Set3"),
        RColorBrewer::brewer.pal(8, "Dark2"))
    return(multiset[1:n])
  }
  else {
    # More that 37, well at least it looks pretty
    return(rainbow(n))
  }
}


ecm <- ExperimentColorMap(

  # The default is viridis::viridis
  # https://cran.r-project.org/web/packages/viridis/vignettes/intro-to-viridis.html#the-color-scales
  # Setting continous is entirely a matter of taste
  # Some find magma easier to read than viridis

  all_continuous = list(
    assays  = viridis::magma,
    colData = viridis::magma,
    rowData = viridis::magma
  ),
  all_discrete = list(
    colData = categorical_color_fun,
    rowData = categorical_color_fun
  )
)


# These options are all sce-contents agnostic.
initial_plots <- c(

  # Show umap with clusters by default
  ReducedDimensionPlot(
                   DataBoxOpen=TRUE,
                   ColorBy="Column data",
                   VisualBoxOpen=TRUE,
                   PanelWidth=6L),

  # Show gene expression plot separated (and coloured) by cluster, by default.
  FeatureAssayPlot(XAxis = "Column data",
                   DataBoxOpen=TRUE,
                   VisualBoxOpen=TRUE,
                   ColorBy="Column data",
                   PanelWidth=6L
                   ),
  # Gene list is better wide
  RowDataTable(PanelWidth=12L),

  # For cell level observations (QC.)
  ColumnDataPlot(PanelWidth=6L,
                 DataBoxOpen=TRUE,
                 VisualBoxOpen=TRUE )
)

app <- iSEE(sce,
            colormap=ecm,
            initial=initial_plots)
"""



#!/usr/bin/env python
# Dan Blankenberg
"""
Reads a list of intervals and a maf. Outputs a new set of intervals with statistics appended.
"""
from __future__ import print_function

import sys

import bx.intervals.io
from bx.bitset import BitSet

from galaxy.tools.util import maf_utilities


def __main__():
    maf_source_type = sys.argv.pop(1)
    input_maf_filename = sys.argv[1].strip()
    input_interval_filename = sys.argv[2].strip()
    output_filename = sys.argv[3].strip()
    dbkey = sys.argv[4].strip()
    try:
        chr_col = int(sys.argv[5].strip()) - 1
        start_col = int(sys.argv[6].strip()) - 1
        end_col = int(sys.argv[7].strip()) - 1
    except Exception:
        print(
            "You appear to be missing metadata. You can specify your metadata by clicking on the pencil icon associated with your interval file.",
            file=sys.stderr,
        )
        sys.exit()
    summary = sys.argv[8].strip()
    if summary.lower() == "true":
        summary = True
    else:
        summary = False

    mafIndexFile = "%s/maf_index.loc" % sys.argv[9]
    try:
        maf_index_filename = sys.argv[10].strip()
    except Exception:
        maf_index_filename = None
    index = index_filename = None
    if maf_source_type == "user":
        # index maf for use here
        index, index_filename = maf_utilities.open_or_build_maf_index(
            input_maf_filename, maf_index_filename, species=[dbkey]
        )
        if index is None:
            print("Your MAF file appears to be malformed.", file=sys.stderr)
            sys.exit()
    elif maf_source_type == "cached":
        # access existing indexes
        index = maf_utilities.maf_index_by_uid(input_maf_filename, mafIndexFile)
        if index is None:
            print("The MAF source specified (%s) appears to be invalid." % (input_maf_filename), file=sys.stderr)
            sys.exit()
    else:
        print("Invalid source type specified: %s" % maf_source_type, file=sys.stdout)
        sys.exit()

    out = open(output_filename, "w")

    num_region = None
    num_bad_region = 0
    species_summary = {}
    total_length = 0
    # loop through interval file
    for num_region, region in enumerate(  # noqa: B007
        bx.intervals.io.NiceReaderWrapper(
            open(input_interval_filename, "r"),
            chrom_col=chr_col,
            start_col=start_col,
            end_col=end_col,
            fix_strand=True,
            return_header=False,
            return_comments=False,
        )
    ):
        src = "%s.%s" % (dbkey, region.chrom)
        region_length = region.end - region.start
        if region_length < 1:
            num_bad_region += 1
            continue
        total_length += region_length
        coverage = {dbkey: BitSet(region_length)}

        for block in index.get_as_iterator(src, region.start, region.end):
            for spec in maf_utilities.get_species_in_block(block):
                if spec not in coverage:
                    coverage[spec] = BitSet(region_length)
            for splitted_block in maf_utilities.iter_blocks_split_by_species(block):
                if maf_utilities.component_overlaps_region(splitted_block.get_component_by_src(src), region):
                    # need to chop and orient the splitted_block
                    oriented_chopped_splitted_block = maf_utilities.orient_block_by_region(
                        maf_utilities.chop_block_by_region(splitted_block, src, region), src, region, force_strand="+"
                    )
                    start_offset, alignment = maf_utilities.reduce_block_by_primary_genome(
                        oriented_chopped_splitted_block, dbkey, region.chrom, region.start
                    )
                    for i in range(len(alignment[dbkey])):
                        for spec, text in alignment.items():
                            if text[i] != "-":
                                coverage[spec].set(start_offset + i)
        if summary:
            # record summary
            for key in coverage.keys():
                if key not in species_summary:
                    species_summary[key] = 0
                species_summary[key] = species_summary[key] + coverage[key].count_range()
        else:
            # print coverage for interval
            coverage_sum = coverage[dbkey].count_range()
            out.write(
                "%s\t%s\t%s\t%s\n" % ("\t".join(region.fields), dbkey, coverage_sum, region_length - coverage_sum)
            )
            keys = list(coverage.keys())
            keys.remove(dbkey)
            keys.sort()
            for key in keys:
                coverage_sum = coverage[key].count_range()
                out.write(
                    "%s\t%s\t%s\t%s\n" % ("\t".join(region.fields), key, coverage_sum, region_length - coverage_sum)
                )
    if summary:
        out.write("#species\tnucleotides\tcoverage\n")
        for spec in species_summary:
            out.write("%s\t%s\t%.4f\n" % (spec, species_summary[spec], float(species_summary[spec]) / total_length))
    out.close()
    if num_region is not None:
        print("%i regions were processed with a total length of %i." % (num_region + 1, total_length))
    if num_bad_region:
        print("%i regions were invalid." % (num_bad_region))
    maf_utilities.remove_temp_index_file(index_filename)


if __name__ == "__main__":
    __main__()



from galaxy.util.commands import which


def find_engine(config):
    nodejs_path = getattr(config, "nodejs_path", None)
    if nodejs_path is None:
        nodejs_path = which("nodejs") or which("node") or None
    if nodejs_path is None:
        raise Exception("nodejs or node not found on PATH")
    return nodejs_path



#!/usr/bin/env python
from __future__ import print_function

import sys
import textwrap

from Bio import GenBank

assert sys.version_info[:2] >= (2, 6)


def make_fasta(rec):
    """Creates fasta format from a record"""
    gi = rec.annotations.get("gi", "")
    org = rec.annotations.get("organism", "")
    date = rec.annotations.get("date", "")
    head = ">gi:%s, id:%s, org:%s, date:%s\n" % (gi, rec.id, org, date)
    body = "\n".join(textwrap.wrap(rec.seq.data, width=80))
    return head, body


if __name__ == "__main__":
    mode = sys.argv[1]
    text = sys.argv[2]
    output_file = sys.argv[3]

    print("Searching for %s <br>" % text)

    # check if inputs are all numbers
    try:
        gi_list = text.split()
        [int(_) for _ in gi_list]
    except ValueError:
        gi_list = GenBank.search_for(text, max_ids=10)

    fp = open(output_file, "wt")
    record_parser = GenBank.FeatureParser()
    ncbi_dict = GenBank.NCBIDictionary(mode, "genbank", parser=record_parser)
    for gid in gi_list:
        res = ncbi_dict[gid]
        head, body = make_fasta(res)
        fp.write(head + body + "\n")
        print(head)
    fp.close()



#!/usr/bin/env python
from __future__ import print_function

import re
import sys
import tempfile

try:
    from rpy2.rpy_classic import (
        BASIC_CONVERSION,
        NO_CONVERSION,
        r,
        RException,
        set_default_mode,
    )
except ImportError:
    # RPy isn't maintained, and doesn't work with R>3.0, use it as a fallback
    from rpy import (
        BASIC_CONVERSION,
        NO_CONVERSION,
        r,
        RException,
        set_default_mode,
    )


def stop_err(msg):
    sys.stderr.write(msg)
    sys.exit(1)


def S3_METHODS(all="key"):
    Group_Math = [
        "abs",
        "sign",
        "sqrt",
        "floor",
        "ceiling",
        "trunc",
        "round",
        "signif",
        "exp",
        "log",
        "cos",
        "sin",
        "tan",
        "acos",
        "asin",
        "atan",
        "cosh",
        "sinh",
        "tanh",
        "acosh",
        "asinh",
        "atanh",
        "lgamma",
        "gamma",
        "gammaCody",
        "digamma",
        "trigamma",
        "cumsum",
        "cumprod",
        "cummax",
        "cummin",
        "c",
    ]
    Group_Ops = [
        "+",
        "-",
        "*",
        "/",
        "^",
        "%%",
        "%/%",
        "&",
        "|",
        "!",
        "==",
        "!=",
        "<",
        "<=",
        ">=",
        ">",
        "(",
        ")",
        "~",
        ",",
    ]
    if all == "key":
        return {"Math": Group_Math, "Ops": Group_Ops}


def main():
    try:
        datafile = sys.argv[1]
        outfile_name = sys.argv[2]
        expression = sys.argv[3]
    except Exception:
        stop_err("Usage: python gsummary.py input_file ouput_file expression")

    math_allowed = S3_METHODS()["Math"]
    ops_allowed = S3_METHODS()["Ops"]

    # Check for invalid expressions
    for word in re.compile("[a-zA-Z]+").findall(expression):
        if word and word not in math_allowed:
            stop_err("Invalid expression '%s': term '%s' is not recognized or allowed" % (expression, word))
    symbols = set()
    for symbol in re.compile(r"[^a-z0-9\s]+").findall(expression):
        if symbol and symbol not in ops_allowed:
            stop_err("Invalid expression '%s': operator '%s' is not recognized or allowed" % (expression, symbol))
        else:
            symbols.add(symbol)
    if len(symbols) == 1 and "," in symbols:
        # User may have entered a comma-separated list r_data_frame columns
        stop_err("Invalid columns '%s': this tool requires a single column or expression" % expression)

    # Find all column references in the expression
    cols = []
    for col in re.compile("c[0-9]+").findall(expression):
        try:
            cols.append(int(col[1:]) - 1)
        except Exception:
            pass

    tmp_file = tempfile.NamedTemporaryFile("w+")
    # Write the R header row to the temporary file
    hdr_str = "\t".join("c%s" % str(col + 1) for col in cols)
    tmp_file.write("%s\n" % hdr_str)
    skipped_lines = 0
    first_invalid_line = 0
    i = 0
    for i, line in enumerate(open(datafile)):
        line = line.rstrip("\r\n")
        if line and not line.startswith("#"):
            valid = True
            fields = line.split("\t")
            # Write the R data row to the temporary file
            for col in cols:
                try:
                    float(fields[col])
                except Exception:
                    skipped_lines += 1
                    if not first_invalid_line:
                        first_invalid_line = i + 1
                    valid = False
                    break
            if valid:
                data_str = "\t".join(fields[col] for col in cols)
                tmp_file.write("%s\n" % data_str)
    tmp_file.flush()

    if skipped_lines == i + 1:
        stop_err(
            "Invalid column or column data values invalid for computation.  See tool tips and syntax for data requirements."
        )
    else:
        # summary function and return labels
        set_default_mode(NO_CONVERSION)
        summary_func = r(
            "function( x ) { c( sum=sum( as.numeric( x ), na.rm=T ), mean=mean( as.numeric( x ), na.rm=T ), stdev=sd( as.numeric( x ), na.rm=T ), quantile( as.numeric( x ), na.rm=TRUE ) ) }"
        )
        headings = ["sum", "mean", "stdev", "0%", "25%", "50%", "75%", "100%"]
        headings_str = "\t".join(headings)

        r_data_frame = r.read_table(tmp_file.name, header=True, sep="\t")

        outfile = open(outfile_name, "w")

        for col in re.compile("c[0-9]+").findall(expression):
            r.assign(col, r["$"](r_data_frame, col))
        try:
            summary = summary_func(r(expression))
        except RException as s:
            outfile.close()
            stop_err("Computation resulted in the following error: %s" % str(s))
        summary = summary.as_py(BASIC_CONVERSION)
        outfile.write("#%s\n" % headings_str)
        if type(summary) is dict:
            # using rpy
            outfile.write("%s\n" % "\t".join("%g" % summary[k] for k in headings))
        else:
            # using rpy2
            outfile.write("%s\n" % "\t".join("%g" % k for k in summary))
        outfile.close()

        if skipped_lines:
            print(
                "Skipped %d invalid lines beginning with line #%d.  See tool tips for data requirements."
                % (skipped_lines, first_invalid_line)
            )


if __name__ == "__main__":
    main()



#!/usr/bin/env python
from __future__ import print_function

import sys

from galaxy.datatypes.util.gff_util import parse_gff_attributes


def get_bed_line(chrom, name, strand, blocks):
    """Returns a BED line for given data."""

    if len(blocks) == 1:
        # Use simple BED format if there is only a single block:
        #   chrom, chromStart, chromEnd, name, score, strand
        #
        start, end = blocks[0]
        return "%s\t%i\t%i\t%s\t0\t%s\n" % (chrom, start, end, name, strand)

    #
    # Build lists for transcript blocks' starts, sizes.
    #

    # Get transcript start, end.
    t_start = sys.maxsize
    t_end = -1
    for block_start, block_end in blocks:
        if block_start < t_start:
            t_start = block_start
        if block_end > t_end:
            t_end = block_end

    # Get block starts, sizes.
    block_starts = []
    block_sizes = []
    for block_start, block_end in blocks:
        block_starts.append(str(block_start - t_start))
        block_sizes.append(str(block_end - block_start))

    #
    # Create BED entry.
    # Bed format: chrom, chromStart, chromEnd, name, score, strand, \
    #               thickStart, thickEnd, itemRgb, blockCount, blockSizes, blockStarts
    #
    # Render complete feature with thick blocks. There's no clear way to do this unless
    # we analyze the block names, but making everything thick makes more sense than
    # making everything thin.
    #
    return "%s\t%i\t%i\t%s\t0\t%s\t%i\t%i\t0\t%i\t%s\t%s\n" % (
        chrom,
        t_start,
        t_end,
        name,
        strand,
        t_start,
        t_end,
        len(block_starts),
        ",".join(block_sizes),
        ",".join(block_starts),
    )


def __main__():
    input_name = sys.argv[1]
    output_name = sys.argv[2]
    skipped_lines = 0
    first_skipped_line = 0
    i = 0
    cur_transcript_chrome = None
    cur_transcript_id = None
    cur_transcript_strand = None
    cur_transcripts_blocks = []  # (start, end) for each block.
    with open(output_name, "w") as out, open(input_name) as in_fh:
        for i, line in enumerate(in_fh):
            line = line.rstrip("\r\n")
            if line and not line.startswith("#"):
                try:
                    # GFF format: chrom source, name, chromStart, chromEnd, score, strand, attributes
                    elems = line.split("\t")
                    start = str(int(elems[3]) - 1)
                    coords = [int(start), int(elems[4])]
                    strand = elems[6]
                    if strand not in ["+", "-"]:
                        strand = "+"
                    attributes = parse_gff_attributes(elems[8])
                    t_id = attributes.get("transcript_id", None)

                    if not t_id:
                        #
                        # No transcript ID, so write last transcript and write current line as its own line.
                        #

                        # Write previous transcript.
                        if cur_transcript_id:
                            # Write BED entry.
                            out.write(
                                get_bed_line(
                                    cur_transcript_chrome,
                                    cur_transcript_id,
                                    cur_transcript_strand,
                                    cur_transcripts_blocks,
                                )
                            )

                        # Replace any spaces in the name with underscores so UCSC will not complain.
                        name = elems[2].replace(" ", "_")
                        out.write(get_bed_line(elems[0], name, strand, [coords]))
                        continue

                    # There is a transcript ID, so process line at transcript level.
                    if t_id == cur_transcript_id:
                        # Line is element of transcript and will be a block in the BED entry.
                        cur_transcripts_blocks.append(coords)
                        continue

                    #
                    # Line is part of new transcript; write previous transcript and start
                    # new transcript.
                    #

                    # Write previous transcript.
                    if cur_transcript_id:
                        # Write BED entry.
                        out.write(
                            get_bed_line(
                                cur_transcript_chrome, cur_transcript_id, cur_transcript_strand, cur_transcripts_blocks
                            )
                        )

                    # Start new transcript.
                    cur_transcript_chrome = elems[0]
                    cur_transcript_id = t_id
                    cur_transcript_strand = strand
                    cur_transcripts_blocks = []
                    cur_transcripts_blocks.append(coords)
                except Exception:
                    skipped_lines += 1
                    if not first_skipped_line:
                        first_skipped_line = i + 1
            else:
                skipped_lines += 1
                if not first_skipped_line:
                    first_skipped_line = i + 1

        # Write last transcript.
        if cur_transcript_id:
            # Write BED entry.
            out.write(
                get_bed_line(cur_transcript_chrome, cur_transcript_id, cur_transcript_strand, cur_transcripts_blocks)
            )
    info_msg = "%i lines converted to BED.  " % (i + 1 - skipped_lines)
    if skipped_lines > 0:
        info_msg += "Skipped %d blank/comment/invalid lines starting with line #%d." % (
            skipped_lines,
            first_skipped_line,
        )
    print(info_msg)


if __name__ == "__main__":
    __main__()



from ast import (
    Module,
    parse,
    walk,
)

AST_NODE_TYPE_ALLOWLIST = [
    "Expr",
    "Load",
    "Str",
    "Num",
    "BoolOp",
    "Compare",
    "And",
    "Eq",
    "NotEq",
    "Or",
    "GtE",
    "LtE",
    "Lt",
    "Gt",
    "BinOp",
    "Add",
    "Div",
    "Sub",
    "Mult",
    "Mod",
    "Pow",
    "LShift",
    "GShift",
    "BitAnd",
    "BitOr",
    "BitXor",
    "UnaryOp",
    "Invert",
    "Not",
    "NotIn",
    "In",
    "Is",
    "IsNot",
    "List",
    "Index",
    "Subscript",
    "Constant",
    # Further checks
    "Name",
    "Call",
    "Attribute",
]


BUILTIN_AND_MATH_FUNCTIONS = "abs|all|any|bin|chr|cmp|complex|divmod|float|hex|int|len|long|max|min|oct|ord|pow|range|reversed|round|sorted|str|sum|type|unichr|unicode|log|exp|sqrt|ceil|floor".split(
    "|"
)
STRING_AND_LIST_METHODS = [name for name in dir("") + dir([]) if not name.startswith("_")]
VALID_FUNCTIONS = BUILTIN_AND_MATH_FUNCTIONS + STRING_AND_LIST_METHODS


def _check_name(ast_node, allowed_variables=None):
    if allowed_variables is None:
        allowed_variables = []
    name = ast_node.id
    return name in (VALID_FUNCTIONS + allowed_variables)


def _check_attribute(ast_node):
    attribute_name = ast_node.attr
    if attribute_name not in STRING_AND_LIST_METHODS:
        return False
    return True


def _check_call(ast_node):
    # If we are calling a function or method, it better be a math,
    # string or list function.
    ast_func = ast_node.func
    ast_func_class = ast_func.__class__.__name__
    if ast_func_class == "Name":
        if ast_func.id not in BUILTIN_AND_MATH_FUNCTIONS:
            return False
    elif ast_func_class == "Attribute":
        if not _check_attribute(ast_func):
            return False
    else:
        return False

    return True


def _check_expression(text, allowed_variables=None):
    """

    >>> allowed_variables = ["c1", "c2", "c3", "c4", "c5"]
    >>> _check_expression("c1", allowed_variables)
    True
    >>> _check_expression("eval('1+1')", allowed_variables)
    False
    >>> _check_expression("import sys", allowed_variables)
    False
    >>> _check_expression("[].__str__", allowed_variables)
    False
    >>> _check_expression("__builtins__", allowed_variables)
    False
    >>> _check_expression("'x' in globals", allowed_variables)
    False
    >>> _check_expression("'x' in [1,2,3]", allowed_variables)
    True
    >>> _check_expression("c3=='chr1' and c5>5", allowed_variables)
    True
    >>> _check_expression("c3=='chr1' and d5>5", allowed_variables)  # Invalid d5 reference
    False
    >>> _check_expression("c3=='chr1' and c5>5 or exec", allowed_variables)
    False
    >>> _check_expression("type(c1) != type(1)", allowed_variables)
    True
    >>> _check_expression("c1.split(',')[1] == '1'", allowed_variables)
    True
    >>> _check_expression("exec 1", allowed_variables)
    False
    >>> _check_expression("str(c2) in [\\\"a\\\",\\\"b\\\"]", allowed_variables)
    True
    """
    if allowed_variables is None:
        allowed_variables = []
    try:
        module = parse(text)
    except SyntaxError:
        return False

    if not isinstance(module, Module):
        return False
    statements = module.body
    if not len(statements) == 1:
        return False
    expression = statements[0]
    if expression.__class__.__name__ != "Expr":
        return False

    for ast_node in walk(expression):
        ast_node_class = ast_node.__class__.__name__

        # Toss out everything that is not a "simple" expression,
        # imports, error handling, etc...
        if ast_node_class not in AST_NODE_TYPE_ALLOWLIST:
            return False

        # White-list more potentially dangerous types AST elements.
        if ast_node_class == "Name":
            # In order to prevent loading 'exec', 'eval', etc...
            # put string restriction on names allowed.
            if not _check_name(ast_node, allowed_variables):
                return False
        # Check only valid, white-listed functions are called.
        elif ast_node_class == "Call":
            if not _check_call(ast_node):
                return False
        # Check only valid, white-listed attributes are accessed
        elif ast_node_class == "Attribute":
            if not _check_attribute(ast_node):
                return False

    return True


def safe_eval(expression, variables):
    """

    >>> safe_eval("moo", {"moo": 5})
    5
    >>> exception_thrown = False
    >>> try: safe_eval("moo", {"cow": 5})
    ... except Exception as e: exception_thrown = True
    >>> exception_thrown
    True
    """
    if not _check_expression(expression, allowed_variables=list(variables.keys())):
        raise Exception(f"Invalid expression [{expression}], only a very simple subset of Python is allowed.")
    return eval(expression, globals(), variables)



"""
Manager and Serializers for Visualizations.

Visualizations are saved configurations/variables used to
reproduce a specific view in a Galaxy visualization.
"""
import logging

from galaxy import model
from galaxy.managers import sharable
from galaxy.structured_app import MinimalManagerApp

log = logging.getLogger(__name__)


class VisualizationManager(sharable.SharableModelManager):
    """
    Handle operations outside and between visualizations and other models.
    """

    # TODO: revisions

    model_class = model.Visualization
    foreign_key_name = "visualization"
    user_share_model = model.VisualizationUserShareAssociation

    tag_assoc = model.VisualizationTagAssociation
    annotation_assoc = model.VisualizationAnnotationAssociation
    rating_assoc = model.VisualizationRatingAssociation

    # def copy( self, trans, visualization, user, **kwargs ):
    #    """
    #    """
    #    pass


class VisualizationSerializer(sharable.SharableModelSerializer):
    """
    Interface/service object for serializing visualizations into dictionaries.
    """

    model_manager_class = VisualizationManager
    SINGLE_CHAR_ABBR = "v"

    def __init__(self, app: MinimalManagerApp):
        super().__init__(app)
        self.visualization_manager = self.manager

        self.default_view = "summary"
        self.add_view("summary", [])
        self.add_view("detailed", [])

    def add_serializers(self):
        super().add_serializers()
        self.serializers.update({})


class VisualizationDeserializer(sharable.SharableModelDeserializer):
    """
    Interface/service object for validating and deserializing
    dictionaries into visualizations.
    """

    model_manager_class = VisualizationManager

    def __init__(self, app):
        super().__init__(app)
        self.visualization_manager = self.manager

    def add_deserializers(self):
        super().add_deserializers()
        self.deserializers.update({})
        self.deserializable_keyset.update(self.deserializers.keys())
